# cvode._jactimes.py

from __future__ import annotations
from typing import Callable


class CVODEJacTimes:
    """Jacobian-vector product."""

    __slots__ = ('setupfn', 'solvefn')

    def __init__(self, setupfn: Callable | None, solvefn: Callable) -> None:
        """
        Wrapper for passing Jacobian-vector product functions to CVODE. The
        Jacobian-vector product interface is only supported by iterative solvers
        (gmres, bicgstab, tfqmr).

        Parameters
        ----------
        setupfn : Callable or None
            A function to setup data before solving the Jacobian-vector product.
            Use None if not needed. The required signature is in the notes.
        solvefn : Callable
            A function that solves for the Jacobian-vector product ``J*v`` (or
            an approximation to it). The required signature is in the notes.

        Raises
        ------
        TypeError
            'setupfn' must be type Callable or None.
        TypeError
            'solvefn' must be type Callable.

        Notes
        -----
        The solve and setup functions require specific function signatures. For
        'solvefn' use ``f(t, y, yp, v, Jv[, userdata])``. Any return values are
        ignored. Instead, the function should fill the pre-allocated memory for
        'Jv' (a 1D array) with the solution (or approximation) to ``J*v``. Use
        ``[:]`` to fill the array rather than overwriting it. For example,
        ``Jv[:] = f(...)`` is correct whereas ``Jv = f(...)`` is not. The user
        is responsible for managing their own Jacobian data if needed. 'setupfn'
        can be used to help setup any needed values/data so that 'solvefn' is
        not overly complex.

        'setupfn' requires the signature ``f(t, y, yp[, userdata])``. As with
        'solvefn', any return values are ignored. However, you can use the
        function to define global variables or add them to ``userdata`` so they
        can be passed to 'solvefn'. The order of function calls is always
        'rhsfn' -> 'setupfn' -> 'solvefn' for each integration step.

        """

        if setupfn is None:
            pass
        elif not isinstance(setupfn, Callable):
            raise TypeError("'setupfn' must be type Callable.")

        if not isinstance(solvefn, Callable):
            raise TypeError("'solvefn' must be type Callable.")

        self.setupfn = setupfn
        self.solvefn = solvefn



# c_nvector.pxd

from .c_sundials cimport *  # Access to types

# nvector_serial.h
cdef extern from "nvector/nvector_serial.h":    
    N_Vector N_VNew_Serial(sunindextype vec_length, SUNContext ctx)
    sunrealtype* N_VGetArrayPointer(N_Vector v)


# _cy_common.pxd

# Dependencies
cimport numpy as np

# Extern cdef headers
from .c_sundials cimport *  # Access to C types

# Convert between N_Vector and numpy array
cdef svec2np(N_Vector nvec, np.ndarray[DTYPE_t, ndim=1] np_array)
cdef np2svec(np.ndarray[DTYPE_t, ndim=1] np_array, N_Vector nvec)

# Convert between sunrealtype* and numpy array
cdef ptr2np(sunrealtype* nv_ptr, np.ndarray[DTYPE_t, ndim=1] np_array)
cdef np2ptr(np.ndarray[DTYPE_t, ndim=1] np_array, sunrealtype* nv_ptr)

# Fill SUNMatrrix with values from 2D numpy array
cdef np2smat(np.ndarray np_A, SUNMatrix smat, object sparsity)



# c_ida.pxd

from .c_sundials cimport *

# ida.h
cdef extern from "ida/ida.h":

    # user-supplied functions
    ctypedef int (*IDAResFn)(sunrealtype t, N_Vector yy, N_Vector yp, N_Vector rr, void* data) except? -1
    ctypedef int (*IDARootFn)(sunrealtype t, N_Vector yy, N_Vector yp, sunrealtype* ee, void* data) except? -1

    # itask
    int IDA_NORMAL
    int IDA_ONE_STEP

    # icopt
    int IDA_YA_YDP_INIT
    int IDA_Y_INIT

    # return values
    int IDA_SUCCESS
    int IDA_TSTOP_RETURN
    int IDA_ROOT_RETURN
    
    # initialization functions
    void* IDACreate(SUNContext ctx)
    int IDAInit(void* mem, IDAResFn resfn, sunrealtype t0, N_Vector y0, N_Vector yp0)
    int IDAReInit(void* mem, sunrealtype t0, N_Vector y0, N_Vector yp0)

    # tolerance input functions
    int IDASStolerances(void* mem, sunrealtype rtol, sunrealtype atol)
    int IDASVtolerances(void* mem, sunrealtype rtol, N_Vector atol)

    # initial condition calculation function
    int IDACalcIC(void* mem, int ic_opt, sunrealtype ic_t0)

    # optional input functions
    int IDASetUserData(void* mem, void* data)
    int IDASetMaxOrd(void* mem, int max_order)
    int IDASetMaxNumSteps(void* mem, long int max_num_steps)
    int IDASetInitStep(void* mem, sunrealtype first_step)
    int IDASetMaxStep(void* mem, sunrealtype max_step)
    int IDASetMinStep(void* mem, sunrealtype min_step)
    int IDASetStopTime(void* mem, sunrealtype tstop)
    int IDAClearStopTime(void* mem)
    int IDASetId(void* mem, N_Vector algidx)
    int IDASetConstraints(void* mem, N_Vector constraints)

    # nonlinear solver input functions
    int IDASetMaxConvFails(void* mem, int max_conv_fails)
    int IDASetMaxNonlinIters(void* mem, int max_nonlin_iters)

    # rootfinding initialization function
    int IDARootInit(void* mem, int nrtfn, IDARootFn eventsfn)

    # rootfinding optional input functions
    int IDASetRootDirection(void* mem, int* direction)

    # solver function
    int IDASolve(void* mem, sunrealtype tend, sunrealtype* tret, N_Vector yret,
                 N_Vector ypret, int itask)
    
    # optional output functions
    int IDAGetConsistentIC(void* mem, N_Vector yy0_mod, N_Vector yp0_mod)
    int IDAGetRootInfo(void* mem, int* rootsfound)
    int IDAGetNumResEvals(void* mem, long int* nrevals)
    int IDAGetCurrentStep(void* mem, sunrealtype* hcur)
    
    # free functions
    void IDAFree(void** mem)

# ida_ls.h
cdef extern from "ida/ida_ls.h":

    # user-supplied functions
    ctypedef int (*IDALsJacFn)(
        sunrealtype t, sunrealtype cj, N_Vector yy, N_Vector yp, N_Vector rr,
        SUNMatrix JJ, void* data, N_Vector tmp1, N_Vector tmp2, N_Vector tmp3) except? -1

    ctypedef int (*IDALsPrecSetupFn)(
        sunrealtype tt, N_Vector yy, N_Vector yp, N_Vector rr, sunrealtype cj,
        void* data) except? -1

    ctypedef int (*IDALsPrecSolveFn)(
        sunrealtype tt, N_Vector yy, N_Vector yp, N_Vector rr, N_Vector rv,
        N_Vector zv, sunrealtype cj, sunrealtype delta, void* data) except? -1

    ctypedef int (*IDALsJacTimesSetupFn)(
        sunrealtype tt, N_Vector yy, N_Vector yp, N_Vector rr, sunrealtype cj,
        void* data) except? -1

    ctypedef int (*IDALsJacTimesVecFn)(
        sunrealtype tt, N_Vector yy, N_Vector yp, N_Vector rr, N_Vector vv, N_Vector Jv,
        sunrealtype cj, void* data, N_Vector tmp1, N_Vector tmp2) except? -1

    # exported functions
    int IDASetLinearSolver(void* mem, SUNLinearSolver LS, SUNMatrix A)

    # optional inputs to LS interface
    int IDASetJacFn(void* mem, IDALsJacFn jacfn)
    int IDASetPreconditioner(void* mem, IDALsPrecSetupFn psetup, IDALsPrecSolveFn psolve)
    int IDASetJacTimes(void* mem, IDALsJacTimesSetupFn jvsetup, IDALsJacTimesVecFn jvsolve)

    # optional outputs from LS interface
    int IDAGetNumJacEvals(void* mem, long int* njevals)



"""
scikit-SUNDAE provides Python bindings to `SUNDIALS`_ integrators. The implicit
differential algebraic (IDA) solver and C-based variable-coefficient ordinary
differential equations (CVODE) solver are both included.

The name SUNDAE combines (SUN)DIALS and DAE, which stands for differential
algebraic equations. Solvers specific to DAE problems are not frequently
available in Python. An ordinary differential equation (ODE) solver is also
included for completeness. ODEs can be categorized as a subset of DAEs (i.e.,
DAEs with no algebraic constraints).

.. _SUNDIALS: https://sundials.readthedocs.io

Accessing the documentation
---------------------------
Documentation is accessible via Python's ``help()`` function which prints
docstrings from a package, module, function, class, etc. You can also access
the documentation by visiting the website, hosted on Read the Docs. The website
includes search functionality and more detailed examples.

Acknowledgements
----------------
scikit-SUNDAE was written by researchers at the **National Renewable Energy
Laboratory (NREL)**, primarily to solve physics-based battery models. Modeling
in Python typically allows for rapid development and makes code more shareable.
However, there was an identified gap in Python's numerical computing ecosystem:
the lack of accessible DAE solvers. While ODE solvers are widely available in
many packages, DAE solvers are not as prevalent.

scikit-SUNDAE started out as a lighter-weight and easy-to-install alternative
to `scikits-odes`_, which similarly provides SUNDIALS bindings, but requires
building from source. The goal was to offer a simpler installation process,
with binary distributions that are consistent across major platforms (PyPI and
conda). While scikit-SUNDAE's API was mostly modeled after scikits-odes, we
want to point out that our codebase was written from scratch and that this is
a separate, independent package. During development we prioritized:

1. Using scipy-like output containers
2. Adopting event function APIs like `scipy.integrate`_, with a few exceptions
3. Maintaining and testing builds using SUNDIALS releases on `conda-forge`_
4. Setting up the package for binary distribution

Since scikit-SUNDAE installations may include pre-built SUNDIALS libraries, a
`bundled license file`_ is included with all distributions. The bundled license
acknowledges not only SUNDIALS, but any optional solvers that are linked and
distributed as well (e.g., SuperLU_MT, OpenBLAS, and LAPACK).

.. _scikits-odes: https://scikits-odes.readthedocs.io
.. _conda-forge: https://anaconda.org/conda-forge/sundials
.. _scipy.integrate: https://docs.scipy.org/doc/scipy/reference/integrate.html
.. _bundled license file: https://github.com/NREL/scikit-sundae/blob/main/ \
                          LICENSES_bundled

"""

from ._cy_common import SUNDIALS_VERSION

from . import ida
from . import utils
from . import cvode
from . import jacband

__all__ = ['ida', 'utils', 'cvode', 'jacband', 'SUNDIALS_VERSION']

__version__ = '1.1.0rc1'



# c_sundials.pxd

# Dependencies
cimport numpy as np

# Define float and int types:
# c_config.pxi is created in setup.py. While building the python package, the 
# sundials_config.h header is parsed to determine what precision was used to
# compile the SUNDIALS that is being built against. The settings are saved in
# the pxi file and used here.
include "c_config.pxi"

# sundials_types.h
cdef extern from "sundials/sundials_types.h":
    ctypedef struct _SUNContext:
        pass

    ctypedef int SUNComm
    ctypedef int sunbooleantype
    ctypedef _SUNContext* SUNContext

    ctypedef void (*SUNErrHandlerFn)(int line, const char* func, const char* file,
                                     const char* msg, int err_code, void* err_user_data,
                                     SUNContext ctx) except *

    int SUN_COMM_NULL

# sundials_context.h
cdef extern from "sundials/sundials_context.h":
    int SUNContext_Create(int comm, SUNContext* ctx)
    void SUNContext_Free(SUNContext* ctx)

    int SUNContext_ClearErrHandlers(SUNContext ctx)
    int SUNContext_PushErrHandler(SUNContext ctx, SUNErrHandlerFn err_fn,
                                  void* err_user_data)

# sundials_nvector.h
cdef extern from "sundials/sundials_nvector.h":
    ctypedef struct _N_Vector:
        pass
    ctypedef _N_Vector* N_Vector

    void N_VDestroy(N_Vector v)

# sundials_matrix.h
cdef extern from "sundials/sundials_matrix.h":
    ctypedef struct _SUNMatrix:
        pass 
    ctypedef _SUNMatrix* SUNMatrix

    ctypedef enum SUNMatrix_ID:
        SUNMATRIX_DENSE
        SUNMATRIX_BAND
        SUNMATRIX_SPARSE

    SUNMatrix_ID SUNMatGetID(SUNMatrix A)

    void SUNMatDestroy(SUNMatrix A)

# sundials_linearsolver.h
cdef extern from "sundials/sundials_linearsolver.h":
    ctypedef struct _SUNLinearSolver:
        pass 
    ctypedef _SUNLinearSolver* SUNLinearSolver

    int SUNLinSolFree(SUNLinearSolver LS)

# sundials_iterative.h
cdef extern from "sundials/sundials_iterative.h":
    cdef enum:
        SUN_PREC_NONE
        SUN_PREC_LEFT
        SUN_PREC_RIGHT
        SUN_PREC_BOTH



"""
Bindings for the CVODE solver in SUNDIALS, used for solving systems of ordinary
differential equations (ODE). Features variable-order, variable-step integration
with support for direct and iterative linear solvers, root finding, and more.

"""

from ._solver import CVODE, CVODEResult
from ._precond import CVODEPrecond
from ._jactimes import CVODEJacTimes

__all__ = [
    'CVODE',
    'CVODEResult',
    'CVODEPrecond',
    'CVODEJacTimes',
]



"""
Bindings for the IDA solver in SUNDIALS, used for solving differential-algebraic
equation (DAE) systems. Supports adaptive time-stepping with support for direct
and iterative linear solvers, root finding, and more.

"""

from ._solver import IDA, IDAResult
from ._precond import IDAPrecond
from ._jactimes import IDAJacTimes

__all__ = [
    'IDA',
    'IDAResult',
    'IDAPrecond',
    'IDAJacTimes',
]



# ida._precond.py

from __future__ import annotations
from typing import Callable


class IDAPrecond:
    """Preconditioner wrapper."""

    __slots__ = ('setupfn', 'solvefn', 'side', '_prectype')

    def __init__(self, setupfn: Callable | None, solvefn: Callable) -> None:
        """
        Wrapper for passing preconditioner functions to IDA. Preconditioning is
        only supported by iterative solvers (gmres, bicgstab, tfqmr). IDA only
        supports left preconditioning. Keep this in mind when defining your
        setup and solve functions.

        Parameters
        ----------
        setupfn : Callable or None, optional
            A function to setup data before solving the preconditioned problem.
            Use None if not needed. The required signature is in the notes.
        solvefn : Callable
            A function that solves the preconditioned problem ``P*zvec = rvec``.
            P is a preconditioner matrix approximating the Jacobian, at least
            crudely. The required signature is in the notes.

        Raises
        ------
        TypeError
            'setupfn' must be type Callable or None.
        TypeError
            'solvefn' must be type Callable.

        Notes
        -----
        The solve and setup functions require specific function signatures. For
        'solvefn' use ``f(t, y, yp, res, rvec, zvec, cj, delta[, userdata])``.
        Any return values are ignored. Instead, the function should fill the
        pre-allocated memory for 'zvec' with the solution to the preconditioned
        problem ``P*zvec = rvec``. Don't forget to use ``[:]`` to fill the array
        rather than overwriting it. For example, ``zvec[:] = f(...)`` is correct
        whereas ``zvec = f(...)`` is not.

        Defining a preconditioning matrix is non-trivial and left to the user.
        For IDA, P should at least crudely approximate the Jacobian given by
        ``J = dF_i/dy_j + cj*dF_i/dyp_j`` where ``res = F(t, y, yp)`` is the
        residual function that describes the system of DAEs. If you need extra
        parameters or values, they can be passed via the optional ``userdata``
        argument. For convenience, you can also define the optional 'setupfn'
        to setup any values (e.g., P) before the solve step.

        The 'setupfn' is an optional function that you can use to perform any
        operations needed before solving. The required signature for 'setupfn'
        is ``f(t, y, yp, res, cj[, userdata])``. Any return values are ignored.
        However, you can use the function to define either global variables or
        add them to ``userdata`` so they can be passed to 'solvefn'. Please
        refer to the `SUNDIALS documentation`_ for more information about these
        functions and their input arguments.

        .. _SUNDIALS documentation: https://sundials.readthedocs.io/en/v6.1.1/ \
            ida/Usage/index.html#preconditioner-setup-iterative-linear-solvers

        """

        if setupfn is None:
            pass
        elif not isinstance(setupfn, Callable):
            raise TypeError("'setupfn' must be type Callable.")

        if not isinstance(solvefn, Callable):
            raise TypeError("'solvefn' must be type Callable.")

        self.setupfn = setupfn
        self.solvefn = solvefn
        self.side = 'left'  # IDA only supports left preconditioning



# cvode._precond.py

from __future__ import annotations
from typing import Callable


class CVODEPrecond:
    """Preconditioner wrapper."""

    __slots__ = ('setupfn', 'solvefn', 'side', '_prectype')

    def __init__(self, setupfn: Callable | None, solvefn: Callable,
                 side: str = 'left') -> None:
        """
        Wrapper for passing preconditioner functions to CVODE. Preconditioning
        is only supported by iterative solvers (gmres, bicgstab, tfqmr).

        Parameters
        ----------
        setupfn : Callable or None
            A function to setup data before solving the preconditioned problem.
            Use None if not needed. The required signature is in the notes.
        solvefn : Callable
            A function that solves the preconditioned problem ``P*zvec = rvec``.
            P is a preconditioner matrix approximating ``I - gamma*J``, at least
            crudely. The required signature is in the notes.
        side : {'left', 'right', 'both'}, optional
            The preconditioning type to use. Can be 'left', 'right', or 'both'.
            The default is 'left'.

        Raises
        ------
        TypeError
            'setupfn' must be type Callable or None.
        TypeError
            'solvefn' must be type Callable.
        ValueError
            'side' must be in {'left', 'right', 'both'}.

        Notes
        -----
        The solve and setup functions require specific function signatures. For
        'solvefn' use ``f(t, y, yp, rvec, zvec, gamma, delta, lr[, userdata])``.
        Any return values are ignored. Instead, the function should fill the
        pre-allocated memory for 'zvec' with the solution to the preconditioned
        problem ``P*zvec = rvec``. Don't forget to use ``[:]`` to fill the array
        rather than overwriting it. For example, ``zvec[:] = f(...)`` is correct
        whereas ``zvec = f(...)`` is not. The input ``lr`` is a flag specifying
        whether left (1) or right (2) preconditioning is being used. In cases
        where 'both' is selected for the preconditioning type, 'solvefn' will
        be called twice and you can control the behavior according to this flag.

        Defining a preconditioning matrix is non-trivial and left to the user.
        For CVODE, P should at least crudely approximate ``I - gamma*J`` where
        ``I`` is the identity and ``J = df_i/dy_j`` is the Jacobian defined by
        the system of ODEs ``yp = f(t, y)``. If you need extra parameters or
        values, they can be passed via the optional ``userdata`` argument. For
        convenience, you can also define the optional 'setupfn' to setup any
        values (e.g., P) before the solve step.

        The 'setupfn' is an optional function that you can use to perform any
        operations needed before solving. The required signature for 'setupfn'
        is ``f(t, y, yp, jok, jnew, gamma[, userdata])``. Any return values are
        ignored. However, you can use the function to define global variables or
        add them to ``userdata`` so they can be passed to 'solvefn'. The inputs
        ``jok`` and ``jnew`` are designed to help optimize performance. The
        ``jok`` flag indicates whether Jacobian data from previous calls can be
        reused (``jok = 1``) or if it must be recomputed (``jok = 0``). The
        ``jnew`` input is a single-item list that the user must control to tell
        the solve whether the Jacobian data has been updated (``jnew[0] = 1``)
        or not (``jnew[0] = 0``). An outlined example is given below.

        .. code-block:: python

            def psetupfn(t, y, yp, jok, jnew, gamma, userdata):
                if jok:
                    jnew[0] = 0
                else:
                    jnew[0] = 1
                    JJ = userdata['JJ']
                    JJ[:,:] = approx_jacobian(...)


            def psolvefn(t, y, yp, rvec, zvec, gamma, delta, lr, userdata):
                Pmat = np.eye(y.size) - gamma*userdata['JJ']
                zvec[:] = ...

        If you need additional information about CVODE preconditioners, please
        reference the original `SUNDIALS documentation`_.

        .. _SUNDIALS documentation: https://sundials.readthedocs.io/en/v6.1.1/ \
            cvode/Usage/index.html#preconditioner-solve-iterative-linear-solvers

        """

        if setupfn is None:
            pass
        elif not isinstance(setupfn, Callable):
            raise TypeError("'setupfn' must be type Callable.")

        if not isinstance(solvefn, Callable):
            raise TypeError("'solvefn' must be type Callable.")

        if side not in {'left', 'right', 'both'}:
            raise ValueError("'side' must be in {'left', 'right', 'both'}.")

        self.setupfn = setupfn
        self.solvefn = solvefn
        self.side = side



# c_cvode.pxd

from .c_sundials cimport *  # Access to types

# cvode.h
cdef extern from "cvode/cvode.h":

    # user-supplied functions
    ctypedef int (*CVRhsFn)(sunrealtype t, N_Vector yy, N_Vector yp, void* data) except? -1
    ctypedef int (*CVRootFn)(sunrealtype t, N_Vector yy, sunrealtype* ee, void* data) except? -1

    # imethod
    int CV_ADAMS
    int CV_BDF
    
    # itask
    int CV_NORMAL
    int CV_ONE_STEP

    # return values
    int CV_SUCCESS
    int CV_TSTOP_RETURN
    int CV_ROOT_RETURN
    
    # initialization functions
    void* CVodeCreate(int imethod, SUNContext ctx)
    int CVodeInit(void* mem, CVRhsFn rhsfn, sunrealtype t0, N_Vector y0)
    int CVodeReInit(void* mem, sunrealtype t0, N_Vector y0)

    # tolerance input functions
    int CVodeSStolerances(void* mem, sunrealtype rtol, sunrealtype atol)
    int CVodeSVtolerances(void* mem, sunrealtype rtol, N_Vector atol)

    # optional input functions
    int CVodeSetUserData(void* mem, void* data)
    int CVodeSetMaxOrd(void* mem, int max_order)
    int CVodeSetMaxNumSteps(void* mem, long int max_num_steps)
    int CVodeSetInitStep(void* mem, sunrealtype first_step)
    int CVodeSetMaxStep(void* mem, sunrealtype max_step)
    int CVodeSetMinStep(void* mem, sunrealtype min_step)
    int CVodeSetStopTime(void* mem, sunrealtype tstop)
    int CVodeClearStopTime(void* mem)
    int CVodeSetConstraints(void* mem, N_Vector constraints)

    # nonlinear solver input functions
    int CVodeSetMaxConvFails(void* mem, int max_conv_fails)
    int CVodeSetMaxNonlinIters(void* mem, int max_nonlin_iters)

    # rootfinding initialization function
    int CVodeRootInit(void* mem, int nrtfn, CVRootFn eventsfn)

    # rootfinding optional input functions
    int CVodeSetRootDirection(void* mem, int* direction)

    # solver function
    int CVode(void* mem, sunrealtype tend, N_Vector yret, sunrealtype* tret, 
              int itask)
    
    # optional output functions
    int CVodeGetRootInfo(void* mem, int* rootsfound)
    int CVodeGetNumRhsEvals(void* mem, long int* nrevals)
    
    # free functions
    void CVodeFree(void** mem)

# cvode_ls.h
cdef extern from "cvode/cvode_ls.h":

    # user-supplied functions
    ctypedef int (*CVLsJacFn)(
        sunrealtype tt, N_Vector yy, N_Vector yp, SUNMatrix JJ, void* data,
        N_Vector tmp1, N_Vector tmp2, N_Vector tmp3) except? -1

    ctypedef int (*CVLsPrecSetupFn)(
        sunrealtype tt, N_Vector yy, N_Vector yp, sunbooleantype jok, sunbooleantype* jcurPtr,
        sunrealtype gamma, void* data) except? -1

    ctypedef int (*CVLsPrecSolveFn)(
        sunrealtype tt, N_Vector yy, N_Vector yp, N_Vector rv, N_Vector zv, sunrealtype gamma,
        sunrealtype delta, int lr, void* data) except? -1

    ctypedef int (*CVLsJacTimesSetupFn)(
        sunrealtype tt, N_Vector yy, N_Vector yp, void* data) except? -1

    ctypedef int (*CVLsJacTimesVecFn)(
        N_Vector vv, N_Vector Jv, sunrealtype tt, N_Vector yy, N_Vector yp, void* data,
        N_Vector tmp) except? -1

    # exported functions
    int CVodeSetLinearSolver(void* mem, SUNLinearSolver LS, SUNMatrix A)

    # optional inputs to LS interface
    int CVodeSetJacFn(void* mem, CVLsJacFn jacfn)
    int CVodeSetPreconditioner(void* mem, CVLsPrecSetupFn psetup, CVLsPrecSolveFn psolve)
    int CVodeSetJacTimes(void* mem, CVLsJacTimesSetupFn jvsetup, CVLsJacTimesVecFn jvsolve)

    # optional outputs from LS interface
    int CVodeGetNumJacEvals(void* mem, long int* njevals)



# ida._solver.py

from __future__ import annotations
from typing import Callable, TYPE_CHECKING

from sksundae._cy_ida import IDA as _IDA, IDAResult as _IDAResult

if TYPE_CHECKING:  # pragma: no cover
    from numpy import ndarray


class IDA:
    """SUNDIALS IDA solver."""

    def __init__(self, resfn: Callable, **options) -> None:
        """
        This class wraps the implicit differential algebraic (IDA) solver from
        SUNDIALS [1]_ [2]_. IDA solves both ordinary differential equations
        (ODEs) and differiential agebraic equations (DAEs).

        Parameters
        ----------
        resfn : Callable
            Residual function with signature ``f(t, y, yp, res[, userdata])``.
            See the notes for more information.
        **options : dict, optional
            Keyword arguments to describe the solver options. A full list of
            names, types, descriptions, and defaults is given below.
        userdata : object or None, optional
            Additional data object to supply to all user-defined callables.
            Cannot be None (default) if 'resfn' takes in 5 arguments.
        calc_initcond : {'y0', 'yp0', None}, optional
            Determines whether or not 'y0' or 'yp0' are corrected before the
            first step. Requires 'calc_init_dt' if not None (default).
        calc_init_dt : float, optional
            Step size for initial condition correction. Positive for forward,
            negative for backward integration. Default is 0.01.
        algebraic_idx : array_like[int] or None, optional
            Indices 'i' for 'y[i]' that are for purely algebraic variables. If
            None (default) the problem is assumed to be a pure ODE.
        first_step : float, optional
            The initial step size. The default is 0, which uses an estimated
            value internally determined by SUNDIALS.
        min_step : float, optional
            Minimum allowable step size. The default is 0.
        max_step : float, optional
            Maximum allowable step size. Use 0 (default) for unbounded steps.
        rtol : float, optional
            Relative tolerance. It is recommended to not use values larger than
            1e-3 or smaller than 1e-15. The default is 1e-5.
        atol : float or array_like[float], optional
            Absolute tolerance. A scalar will apply to all variables equally,
            while an array (matching 'y' length) sets specific tolerances for
            eqch variable. The default is 1e-6.
        linsolver : {'dense', 'band', 'sparse', ...}, optional
            Choice of linear solver, defaults to 'dense'. 'band' requires both
            'lband' and 'uband'. 'sparse' uses SuperLU_MT [3]_ and requires
            'sparsity'. When using an iterative method ('gmres', 'bicgstab',
            'tfqmr') the number of Krylov dimensions is set using 'krylov_dim'.
            'lapackdense' and 'lapackband' can also be used as alternatives to
            'dense' and 'band'. They use OpenBLAS-linked LAPACK [4]_ routines,
            but can have noticeable overhead for small (<100) systems.
        lband : int or None, optional
            Lower Jacobian bandwidth. Given a DAE system ``0 = F(t, y, yp)``,
            the Jacobian is ``J = dF_i/dy_j + cj*dF_i/dyp_j``. Required when
            'linsolver' is 'band'. Use zero if no values are below the main
            diagonal. Defaults to None.
        uband : int or None, optional
            Upper Jacobian bandwidth. Required when 'linsolver' is 'band'. Use
            zero if no elements are above the main diagonal. Defaults to None.
        sparsity : 2D np.array or sparse matrix or None, optional
            Jacobian sparsity pattern. Required when 'linsolver' is 'sparse'.
            The shape must be (N, N) where N is the size of the system. Zero
            entries indicate fixed zeros in the Jacobian. If 'jacfn' is None,
            this argument activates a custom Jacobian routine (not part of the
            original SUNDIALS package). The routine works with all direct linear
            solvers but may increase step count. Reduce 'max_step' to help with
            this, if needed. Defaults to None.
        nthreads : int or None, optional
            Number of threads to use with the 'sparse' linear solver. If None
            (default), 1 is used. Use -1 to use all available threads.
        krylov_dim : int or None, optional
            Maximum number of Krylov basis vectors for iterative solvers. Will
            default to 5 if invalid/None when required. Larger values improve
            convergence but increase memory usage. Only applies to the 'gmres',
            'bicgstab', and 'tfqmr' linear solvers.
        max_order : int, optional
            Specifies the maximum order for the linear multistep BDF method.
            The value must be in the range [1, 5]. The default is 5.
        max_num_steps : int, optional
            The maximum number of steps taken by the solver in each attempt to
            reach the next output time. The default is 500.
        max_nonlin_iters : int, optional
            Specifies the maximum number of nonlinear solver iterations in one
            step. The default is 4.
        max_conv_fails : int, optional
            Specifies the max number of nonlinear solver convergence failures
            in one step. The default is 10.
        constraints_idx : array_like[int] or None, optional
            Specifies indices 'i' in the 'y' state variable array for which
            inequality constraints should be applied. Constraint types must be
            specified in 'constraints_type', see below. The default is None.
        constraints_type : array_like[int] or None, optional
            If 'constraints_idx' is not None, then this option must include an
            array of equal length specifying the types of constraints to apply.
            Values should be in ``{-2, -1, 1, 2}`` which apply ``y[i] < 0``,
            ``y[i] <= 0``, ``y[i] >=0,`` and ``y[i] > 0``, respectively. The
            default is None.
        eventsfn : Callable or None, optional
            Events function with signature ``g(t, y, yp, events[, userdata])``.
            If None (default), no events are tracked. See the notes for more
            information. Requires 'num_events' be set when not None.

            The function may also have these optional attributes:

                terminal: list[bool, int], optional
                    Specifies solver behavior for each event. A boolean stops
                    the solver (True) or just records the event (False). An
                    integer stops the solver after than many occurrences. The
                    default is ``[True]*num_events``.
                direction: list[int], optional
                    Determines which event slopes to track: ``-1`` (negative),
                    ``1`` (positive), or ``0`` (both). If not provided the
                    default ``[0]*num_events`` is used.

            You can assign attributes like ``eventsfn.terminal = [True]`` to
            any function in Python, after it has been defined.
        num_events : int, optional
            Number of events to track. The default is 0.
        jacfn : Callable or None, optional
            Jacobian function like ``J(t, y, yp, res, cj, JJ[, userdata])``.
            The function should fill the pre-allocated 2D matrix 'JJ' with the
            values defined by ``JJ[i,j] = dres_i/dy_j + cj*dres_i/dyp_j``. An
            internal finite difference method is applied when None (default).
        precond : IDAPrecond or None, optional
            Preconditioner functions. Only compatible with iterative linear
            solvers. Must be an instance of IDAPrecond if not None (default).
        jactimes : IDAJacTimes or None, optional
            Jacobian-vector product functions. Only compatible with iterative
            linear solvers. Must be an instance of IDAJacTimes when provided.
            Difference quotient approximations are used with iterative solvers
            if None (default).

        Notes
        -----
        Return values from all user-defined function (e.g., 'resfn', 'eventsfn',
        and 'jacfn') are ignored by the solver. Instead the solver directly
        reads from pre-allocated memory. Output arrays (e.g., 'res', 'events',
        and 'JJ') from each user-defined callable should be filled within each
        respective function. When setting values across the entire array/matrix
        at once, don't forget to use ``[:]`` to fill the existing array rather
        than overwriting it. For example, using ``res[:] = F(t, y, yp)`` is
        correct whereas ``res = F(t, y, yp)`` is not.

        When any user-defined function require data outside of their normal
        arguments, you can supply optional 'userdata'. When given, 'userdata'
        must appear in ALL function signatures ('rhsfn', 'eventsfn', 'jacfn',
        'precond', and 'jactimes') even if it is not used in all functions. Of
        course this does not apply to functions that are provided as ``None``.
        Note that 'userdata' only takes up one argument position; however,
        'userdata' can be any Python object. Therefore, to pass more than one
        extra argument you should pack all data into a single tuple, dict,
        dataclass, etc. and pass them all together as 'userdata'. The data can
        be unpacked as needed within the functions.

        References
        ----------
        .. [1] A. C. Hindmarsh, P. N. Brown, K. E. Grant, S. L. Lee, R.
           Serban, D. E. Shumaker, and C. S. Woodward, "SUNDIALS: Suite of
           Nonlinear and Differential/Algebraic Equation Solvers," ACM TOMS,
           2005, DOI: 10.1145/1089014.1089020
        .. [2] D. J. Gardner, D. R. Reynolds, C. S. Woodward, C. J. Balos,
           "Enabling new flexibility in the SUNDIALS suite of nonlinear and
           differential/algebraic equation solvers," ACM TOMS, 2022,
           DOI: 10.1145/3539801
        .. [3] J. W. Demmel, J. R. Gilbert, and X. S. Li, "An Asynchronous
           Parallel Supernodal Algorithm for Sparse Gaussian Elimination,"
           SIMAX, 1999, DOI: 10.1137/S0895479897317685
        .. [4] E. Anderson, Z. Bai, C. Bischof, S. Blackford, J. Demmel, J.
           Dongarra, J. Du Croz, A. Greenbaum, S. Hammarling, A. McKenney, D.
           Sorensen, "LAPACK Users' Guide," Society for Industrial and Applied
           Mathematics, 1999, Philidelphia, PA.

        Examples
        --------
        The following example solves the Robertson problem, which is a classic
        test problem for programs that solve stiff ODEs. A full description of
        the problem is provided by `MATLAB <Rob-Ex_>`_. While initializing the
        solver, ``algebraic_idx=[2]`` specifies ``y[2]`` is purely algebraic,
        and ``calc_initcond='yp0'`` tells the solver to determine the values
        for 'yp0' at 'tspan[0]' before starting to integrate. That is why 'yp0'
        can be initialized as an array of zeros even though plugging in 'y0'
        to the residuals expressions actually gives ``yp0 = [-0.04, 0.04, 0]``.
        The initialization is checked against the correct answer after solving.

        .. _Rob-Ex:
            https://www.mathworks.com/help/matlab/math/
            solve-differential-algebraic-equations-daes.html

        .. code-block:: python

            import numpy as np
            import sksundae as sun
            import matplotlib.pyplot as plt

            def resfn(t, y, yp, res):
                res[0] = yp[0] + 0.04*y[0] - 1e4*y[1]*y[2]
                res[1] = yp[1] - 0.04*y[0] + 1e4*y[1]*y[2] + 3e7*y[1]**2
                res[2] = y[0] + y[1] + y[2] - 1.0

            solver = sun.ida.IDA(resfn, algebraic_idx=[2], calc_initcond='yp0')

            tspan = np.hstack([0, 4*np.logspace(-6, 6)])
            y0 = np.array([1, 0, 0])
            yp0 = np.zeros_like(y0)

            soln = solver.solve(tspan, y0, yp0)
            assert np.allclose(soln.yp[0], [-0.04, 0.04, 0], rtol=1e-3)

            soln.y[:, 1] *= 1e4  # scale y[1] so it is visible in the figure
            plt.semilogx(soln.t, soln.y)
            plt.show()

        """
        self._IDA = _IDA(resfn, **options)

    def init_step(self, t0: float, y0: ndarray, yp0: ndarray) -> IDAResult:
        """
        Initialize the solver.

        This method is called automatically when using 'solve'. However, it
        must be run manually, before the 'step' method, when solving with a
        step-by-step approach.

        Parameters
        ----------
        t0 : float
            Initial value of time.
        y0 : array_like[float], shape(m,)
            State variable values at 't0'. The length must match that of 'yp0'
            and the number of residual equations in 'resfn'.
        yp0 : array_like[float], shape(m,)
            Time derivatives for the 'y0' array, evaluated at 't0'. The length
            and indexing should be consistent with 'y0'.

        Returns
        -------
        :class:`~sksundae.ida.IDAResult`
            Custom output class for IDA solutions. Includes pretty-printing
            consistent with scipy outputs. See the class definition for more
            information.

        Raises
        ------
        MemoryError
            Failed to allocate memory for the IDA solver.
        RuntimeError
            A SUNDIALS function returned NULL or was unsuccessful.
        ValueError
            'y0' and 'yp0' must be the same length.

        """
        return self._IDA.init_step(t0, y0, yp0)

    def step(self, t: float, method='normal', tstop=None) -> IDAResult:
        """
        Return the solution at time 't'.

        Before calling the 'step' method, you must first initialize the solver
        by running 'init_step'.

        Parameters
        ----------
        t : float
            Value of time.
        method : {'normal', 'onestep'}, optional
            Solve method for the current step. When 'normal' (default), output
            is returned at time 't'. If 'onestep', output is returned after one
            internal step toward 't'. Both methods stop at events, if given,
            regardless of how 'eventsfn.terminal' was set.
        tstop : float, optional
            Specifies a hard time constraint for which the solver should not
            pass, regardless of the 'method'. The default is None.

        Returns
        -------
        :class:`~sksundae.ida.IDAResult`
            Custom output class for IDA solutions. Includes pretty-printing
            consistent with scipy outputs. See the class definition for more
            information.

        Raises
        ------
        ValueError
            'method' value is invalid. Must be 'normal' or 'onestep'.
        ValueError
            'init_step' must be run prior to 'step'.

        Notes
        -----
        In general, when solving step by step, times should all be provided in
        either increasing or decreasing order. The solver can output results at
        times taken in the opposite direction of integration if the requested
        time is within the last internal step interval; however, values outside
        this interval will raise errors. Rather than trying to mix forward and
        reverse directions, choose each sequential time step carefully so you
        get all of the values you need.

        SUNDIALS provides a convenient graphic to help users understand how the
        step method and optional 'tstop' affect where the integrator stops. To
        read more, see their documentation `here`_.

        .. _here: https://computing.llnl.gov/projects/sundials/usage-notes

        """
        return self._IDA.step(t, method, tstop)

    def solve(self, tspan: ndarray, y0: ndarray, yp0: ndarray) -> IDAResult:
        """
        Return the solution across 'tspan'.

        Parameters
        ----------
        tspan : array_like[float], shape(n >= 2,)
            Solution time span. If ``len(tspan) == 2``, the solution will be
            saved at internally chosen steps. When ``len(tspan) > 2``, the
            solution saves the output at each specified time.
        y0 : array_like[float], shape(m,)
            State variable values at 'tspan[0]'. The length must match that of
            'yp0' and the number of residual equations in 'resfn'.
        yp0 : array_like[float], shape(m,)
            Time derivatives for the 'y0' array, evaluated at 'tspan[0]'. The
            length and indexing should be consistent with 'y0'.

        Returns
        -------
        :class:`~sksundae.ida.IDAResult`
            Custom output class for IDA solutions. Includes pretty-printing
            consistent with scipy outputs. See the class definition for more
            information.

        Raises
        ------
        ValueError
            'tspan' must be strictly increasing or decreasing.
        ValueError
            'tspan' length must be >= 2.

        """
        return self._IDA.solve(tspan, y0, yp0)


class IDAResult(_IDAResult):
    """Results container."""

    def __init__(self, **kwargs) -> None:
        """
        Inherits from :class:`~sksundae.common.RichResult`. The solution class
        groups output from :class:`IDA` into an object with the fields:

        Parameters
        ----------
        message : str
            Human-readable description of the status value.
        success : bool
            True if the solver was successful (status >= 0). False otherwise.
        status : int
            Reason for the algorithm termination. Negative values correspond
            to errors, and non-negative values to different successful criteria.
        t : ndarray, shape(n,)
            Solution time(s). The dimension depends on the method. Stepwise
            solutions will only have 1 value whereas solutions across a full
            'tspan' will have many.
        y : ndarray, shape(n, m)
            State variable values at each solution time. Rows correspond to
            indices in 't' and columns match indexing from 'y0'.
        yp : ndarray, shape(n, m)
            State variable time derivate values at each solution time. Row
            and column indexing matches 'y'.
        i_events : ndarray, shape(k, num_events) or None
            Provides an array for each detected event 'k' specifying indices
            for which event(s) occurred. ``i_events[k,i] != 0`` if 'events[i]'
            occurred at 't_events[k]'. The sign of 'i_events' indicates the
            direction of zero-crossing:

                * -1 indicates 'events[i]' was decreasing
                * +1 indicates 'events[i]' was increasing

            Output for 'i_events' will be None when either 'eventsfn' was None
            or if no events occurred during the solve.
        t_events : ndarray, shape(k,) or None
            Times at which events occurred or None if 'eventsfn' was None or
            no events were triggered during the solve.
        y_events : ndarray, shape(k, m) or None
            State variable values at each 't_events' value or None. Rows and
            columns correspond to 't_events' and 'y0' indexing, respectively.
        yp_events : ndarray, shape(k, m) or None
            State variable time derivative values at each 't_events' value or
            None. Row and column indexing matches 'y_events'.
        nfev : int
            Number of times that 'resfn' was evaluated.
        njev : int
            Number of times the Jacobian was evaluated, 'jacfn' or internal
            finite difference method.

        Notes
        -----
        Terminal events are appended to the end of 't', 'y', and 'yp'. However,
        if an event was not terminal then it will only appear in '\\*_events'
        outputs and not within the main output arrays.

        'nfev' and 'njev' are cumulative for stepwise solution approaches. The
        values are reset each time 'init_step' is called.

        """
        super().__init__(**kwargs)



"""
General-purpose module for shared utilities across the package.

"""

import numpy as np

__all__ = ['RichResult',]


# RichResult and its formatters are modified copies from scipy._lib._util
class RichResult:
    """Output container with pretty printing."""

    _order_keys = []

    def __init__(self, **kwargs):
        """
        This class is a based off the ``_RichResult`` class in the ``scipy``
        library. It combines a series of formatting functions to make the
        printed 'repr' easy to read. Use this class directly by passing in
        any number of keyword arguments, or use it as a base class to have
        custom classes for different result types.

        Inheriting classes should define the class attribute ``_order_keys``
        which is a list of strings that defines how output fields are sorted
        when an instance is printed.

        Parameters
        ----------
        **kwargs : dict, optional
            User-specified keyword arguments. Any number of arguments can be
            given as input. The class simply stores the key/value pairs, makes
            them accessible as attributes, and provides pretty printing.

        Examples
        --------
        The example below demonstrates how to define the ``_order_keys`` class
        attribute for custom sorting. If arguments are not in the list, they
        are placed at the end based on the order they were given. Note that
        ``_order_keys`` only provides sorting support and that no errors are
        raised if an argument is not present, e.g., ``third`` below.

        .. code-block:: python

            import sksundae as sun

            class CustomResult(sun.utils.RichResult):
                _order_keys = ['first', 'second', 'third',]

            result = CustomResult(second=None, last=None, first=None)
            print(result)

        ``RichResult`` can also be used directly, without any custom sorting.
        Arguments will print based on the order they were input. Instances will
        still have a fully formatted 'repr', including formatted arrays.

        .. code-block:: python

            import numpy as np

            from sksundae.utils import RichResult

            t = np.linspace(0, 1, 1000)
            y = np.random.rand(1000, 5)

            y[0] = np.inf
            y[-1] = np.nan

            result = RichResult(message='Example.', status=0, t=t, y=y)
            print(result)

        After initialization, all key/value pairs are accessible as instance
        attributes.

        .. code-block:: python

            from sksundae.utils import RichResult

            result = RichResult(a=10, b=20, c=30)

            print(result.a*(result.b + result.c))

        """
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self):
        order_keys = getattr(self, '_order_keys')

        def key(item):
            try:
                return order_keys.index(item[0].lower())
            except ValueError:  # item not in list, move to end
                return np.inf

        def sorter(d):
            return sorted(d.items(), key=key)

        if self.__dict__.keys():
            return '\n' + _format_dict(self.__dict__, sorter=sorter) + '\n'
        else:
            return self.__class__.__name__ + '()'


def _indenter(s, n=0):
    """Ensures lines after the first are indented by the specified amount."""

    split = s.split('\n')
    return ('\n' + ' '*n).join(split)


def _format_float_10(x):
    """Returns string representation of floats with exactly ten characters."""

    if np.isposinf(x):
        return '       inf'
    elif np.isneginf(x):
        return '      -inf'
    elif np.isnan(x):
        return '       nan'
    return np.format_float_scientific(x, precision=3, pad_left=2, unique=False)


def _format_dict(d, n=0, mplus=1, sorter=None):
    """Pretty printer for dictionaries."""

    if isinstance(d, dict):
        m = max(map(len, list(d.keys()))) + mplus  # width to print keys

        # indent and format each value... + 2 for ': '
        line_items = []
        for k, v in sorter(d):
            formatted_v = _indenter(_format_dict(v, m+n+2, 0, sorter), m+2)
            line_items.append(k.rjust(m) + ': ' + formatted_v)

        output_str = '\n'.join(line_items)

    else:
        with np.printoptions(linewidth=76-n, edgeitems=2, threshold=12,
                             formatter={'float_kind': _format_float_10}):

            output_str = str(d)

    return output_str



# c_sunlinsol.pxd

from .c_sundials cimport *  # Access to types

# sunlinsol_dense.h
cdef extern from "sunlinsol/sunlinsol_dense.h":
    SUNLinearSolver SUNLinSol_Dense(N_Vector y, SUNMatrix A, SUNContext ctx)

# sunlinsol_band.h
cdef extern from "sunlinsol/sunlinsol_band.h":
    SUNLinearSolver SUNLinSol_Band(N_Vector y, SUNMatrix A, SUNContext ctx)

# sunlinsol_spgmr.h
cdef extern from "sunlinsol/sunlinsol_spgmr.h":
    SUNLinearSolver SUNLinSol_SPGMR(N_Vector y, int pretype, int maxl,
                                    SUNContext ctx)

# sunlinsol_spbcgs.h
cdef extern from "sunlinsol/sunlinsol_spbcgs.h":
    SUNLinearSolver SUNLinSol_SPBCGS(N_Vector y, int pretype, int maxl,
                                     SUNContext ctx)

# sunlinsol_sptfqmr.h
cdef extern from "sunlinsol/sunlinsol_sptfqmr.h":
    SUNLinearSolver SUNLinSol_SPTFQMR(N_Vector y, int pretype, int maxl,
                                      SUNContext ctx)

# sunlinsol_superlumt.h - real or dummy, depending on availability
cdef extern from "./include/superlumt_wrapper.h":
    SUNLinearSolver SUNLinSol_SuperLUMT(N_Vector y, SUNMatrix A, int nthreads,
                                        SUNContext ctx)

# sunlinsol_lapackdense.h - real or dummy, depending on availability
cdef extern from "./include/lapackdense_wrapper.h":
    SUNLinearSolver SUNLinSol_LapackDense(N_Vector y, SUNMatrix A,
                                          SUNContext ctx)

# sunlinsol_lapackband.h - real or dummy, depending on availability
cdef extern from "./include/lapackband_wrapper.h":
    SUNLinearSolver SUNLinSol_LapackBand(N_Vector y, SUNMatrix A,
                                         SUNContext ctx)



"""
Routines for analyzing Jacobian patterns and structures for problems involving
systems of differential-algebraic equations. For example, finding the sparsity
pattern and/or bandwidth.

"""

from __future__ import annotations
from typing import Callable, Any, TYPE_CHECKING

import inspect
from warnings import warn

import numpy as np

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import ndarray
    from scipy.sparse import spmatrix


def _cvode_pattern(rhsfn: Callable, t0: float, y0: ndarray,
                   userdata: Any = None) -> ndarray:
    """Jacobian pattern for CVODE functions. Access via j_pattern()."""

    # wrap rhsfn for cases w/ and w/o userdata
    signature = inspect.signature(rhsfn)

    if len(signature.parameters) == 3:
        def wrapper(t, y, yp): return rhsfn(t, y, yp)
    elif len(signature.parameters) == 4:
        if userdata is None:
            warn("'rhsfn' signature has 4 inputs so 'userdata' is expected,"
                 " but 'userdata=None'. Ensure this was intentional.")

        def wrapper(t, y, yp): return rhsfn(t, y, yp, userdata)
    else:
        raise ValueError("'rhsfn' signature must have either 3 or 4 inputs.")

    # recommended minimum perturbation
    dtype = y0.dtype if np.issubdtype(y0.dtype, np.floating) else np.float64
    uround = np.finfo(dtype).eps
    srur = np.sqrt(uround)

    # perturbed variables
    sign_y = (y0 >= 0).astype(float) * 2 - 1
    y = sign_y * np.maximum(uround, np.abs(y0))

    # initial derivatives
    yp_0 = np.zeros_like(y)
    wrapper(t0, y, yp_0)

    norm = max(srur, np.abs(yp_0).max())
    yp_0 = yp_0 / norm

    # Jacobian pattern
    def j_pattern(j):

        y_store = y[j]
        sign = (y[j] >= 0).astype(float) * 2 - 1

        y[j] += sign * srur * max(1.0, abs(y[j]))

        yp = np.zeros_like(y)
        wrapper(t0, y, yp)
        yp = yp / norm

        y[j] = y_store

        return (yp_0 != yp).astype(float)

    j_cols = [j_pattern(j) for j in range(y.size)]

    return np.column_stack(j_cols)


def _ida_pattern(resfn: Callable, t0: float, y0: ndarray, yp0: ndarray = None,
                 userdata: Any = None) -> ndarray:
    """Jacobian pattern for IDA functions. Access via j_pattern()."""

    # wrap resfn for cases w/ and w/o userdata
    signature = inspect.signature(resfn)

    if len(signature.parameters) == 4:
        def wrapper(t, y, yp, res): return resfn(t, y, yp, res)
    elif len(signature.parameters) == 5:
        if userdata is None:
            warn("'rhsfn' signature has 5 inputs so 'userdata' is expected,"
                 " but 'userdata=None'. Ensure this was intentional.")

        def wrapper(t, y, yp, res): return resfn(t, y, yp, res, userdata)
    else:
        raise ValueError("'rhsfn' signature must have either 4 or 5 inputs.")

    # recommended minimum perturbation
    dtype = y0.dtype if np.issubdtype(y0.dtype, np.floating) else np.float64
    uround = np.finfo(dtype).eps
    srur = np.sqrt(uround)

    # perturbed variables
    sign_y = (y0 >= 0).astype(float) * 2 - 1
    y = sign_y * np.maximum(uround, np.abs(y0))

    sign_yp = (yp0 >= 0).astype(float) * 2 - 1
    yp = sign_yp * np.maximum(uround, np.abs(yp0))

    # initial residuals
    res_0 = np.zeros_like(y)
    wrapper(t0, y, yp, res_0)

    norm = max(srur, np.abs(res_0).max())
    res_0 = res_0 / norm

    # Jacobian pattern
    rng = np.random.default_rng(42)
    rand = rng.random(2)

    def j_pattern(j):

        y_store, yp_store = y[j], yp[j]
        sign = (yp[j] >= 0).astype(float) * 2 - 1

        y[j] += sign * srur * rand[0] * max(1.0, abs(y[j]))
        yp[j] += sign * srur * rand[1] * max(1.0, abs(yp[j]))

        res = np.zeros_like(y)
        wrapper(t0, y, yp, res)
        res = res / norm

        y[j], yp[j] = y_store, yp_store

        return (res_0 != res).astype(float)

    j_cols = [j_pattern(j) for j in range(y.size)]

    return np.column_stack(j_cols)


def j_pattern(rhsfn: Callable, t0: float, y0: ndarray, yp0: ndarray = None,
              userdata: Any = None) -> ndarray:
    """
    Approximate the Jacobian pattern.

    This function uses a numerical Jacobian approximation for ``rhsfn`` about
    the given point to determine the Jacobian pattern. It requires evaluating
    the given function ``N`` times based on the size of ``y0`` so it can be
    slow for large systems.

    Be aware that this routine may return zeros in locations where ones should
    be depending on the evaluation point ``y0`` (and ``yp0``). It is left to
    the user to determine a representative evaluation point that ensures all
    relevant relationships are correctly determined.

    Parameters
    ----------
    rhsfn : Callable
        Right-hand-side function for either an :class:`~sksundae.ida.IDA` or
        :class:`~sksundae.cvode.CVODE` problem. Signatures vary, see the class
        docstrings for more info. The correct routine is automatically selected
        depending on whether or not ``yp0`` is given. IDA requires that ``yp0``
        be given and CVODE expects it to be None.
    t0 : float
        Input time to use in 'rhsfn'.
    y0 : ndarray
        State variables to use in 'rhsfn'.
    yp0 : ndarray or None, optional
        State variable time derivatives to use in 'rhsfn', by default None.
    userdata : Any, optional
        Additional data to pass to 'rhsfn', by default None.

    Returns
    -------
    pattern : 2D np.array
        Jacobian pattern represented by a 2D numpy array with shape (N, N).
        Ones or zeros in the position ``A[i, j]`` mean that function ``F_i``
        either is or is not dependedent on variable ``y_j``, respectively.

    """

    if yp0 is None:
        y0 = np.asarray(y0, dtype=float)
        return _cvode_pattern(rhsfn, t0, y0, userdata)
    else:
        y0 = np.asarray(y0, dtype=float)
        yp0 = np.asarray(yp0, dtype=float)
        return _ida_pattern(rhsfn, t0, y0, yp0, userdata)


def bandwidth(A: ndarray) -> tuple[int]:
    """
    Return half bandwidths of a 2D array.

    Uses the ``scipy.linalg.bandwidth`` function to determine the lower and
    upper bandwidths of a given array. Use in conjunction with ``j_pattern``
    to find the bandwidths of a Jacobian pattern.

    Parameters
    ----------
    A : 2D np.array
        Input array of size (N, M).

    Returns
    -------
    bands : tuple[int]
        A 2-tuple of integers indicating the lower and upper half bandwidths
        of the given matrix. A zero denotes that there are no non-zero elements
        on that side. The full bandwidth is ``lband + uband + 1``.

    """
    from scipy.linalg import bandwidth
    return bandwidth(A)


def reduce_bandwidth(A: ndarray | spmatrix,
                     symmetric: bool = False) -> tuple[ndarray]:
    """
    Find a row/col reordering to reduce bandwidth.

    Uses the Reverse Cuthill-McKee algorithm from ``scipy.sparse.csgraph`` to
    determine an index rearragement for rows and columns of ``A`` that reduce
    the bandwidth.

    Parameters
    ----------
    A : ndarray | spmatrix
        A 2D (n, n) input matrix whose sparsity pattern will be reduced.
    symmetric : bool, optional
        True if input matrix is guaranteed symmetric, otherwise False (default).

    Returns
    -------
    perm : ndarray
        Array of permuted row and column indices such that B = A[perm][:, perm]
        is a maxtrix with reduced bandwidth compared to A.
    inv_perm : ndarray
        The inverse of 'perm' such that B[inv_perm][:, inv_perm] returns the
        original matrix A.

    """
    import scipy.sparse as sp
    from scipy.sparse.csgraph import reverse_cuthill_mckee

    if not sp.issparse(A):
        A = sp.csc_matrix(A)
    else:
        A = A.tocsc()

    perm = reverse_cuthill_mckee(A, symmetric)
    inv_perm = np.argsort(perm)

    return perm, inv_perm



# setup.py

import os
import re
import setuptools

from warnings import warn

import numpy as np

from Cython.Build import cythonize
from packaging.version import Version

MIN_VERSION = Version('7.3.0')
MAX_VERSION = Version('7.6.0')


def find_sundials():
    search_paths = []

    SUNDIALS_PREFIX = os.environ.get('SUNDIALS_PREFIX')
    if SUNDIALS_PREFIX:
        search_paths.append(SUNDIALS_PREFIX)

    CONDA_PREFIX = os.environ.get('CONDA_PREFIX')
    if CONDA_PREFIX:
        search_paths.extend([
            CONDA_PREFIX,
            os.path.join(CONDA_PREFIX, 'Library'),
        ])

    search_paths.extend([
        '/usr',
        '/usr/local',
        'C:/SUNDIALS',
        'C:/Program Files/SUNDIALS',
    ])

    for BASE in search_paths:
        include_dir = os.path.join(BASE, 'include')
        CONFIG_H = os.path.join(include_dir, 'sundials', 'sundials_config.h')
        if os.path.exists(CONFIG_H):
            return BASE, CONFIG_H

    raise FileNotFoundError(
        f"Can't find SUNDIALS installation in any of the {search_paths=}. Set"
        " the environment variable SUNDIALS_PREFIX to the parent directory of"
        " the 'include' and 'lib' directories and retry the installation."
    )


def parse_config_h(file):

    config = {}
    define_rx = re.compile(r"#define\s+(\w+)(?:\s+(.*))?")

    for line in file:

        line = line.strip()
        if define_rx.match(line):

            name, value = define_rx.match(line).group(1, 2)
            if isinstance(value, str):
                value = value.strip('"')

            try:
                config[name] = int(value)
            except (ValueError, TypeError):  # Catch non int and NoneType
                config[name] = value

    return {k: config[k] for k in sorted(config)}


def get_extensions():

    # Parse sundials_config.h
    BASE, CONFIG_H = find_sundials()
    with open(CONFIG_H, 'r') as f:
        config = parse_config_h(f)

    print("\n\n")
    for k, v in config.items():
        print(k, v)
    print("\n\n")

    # Write pxi files for C and Python to match types to sundials_config.h
    SUNDIALS_VERSION = config.get('SUNDIALS_VERSION')
    if not (MIN_VERSION <= Version(SUNDIALS_VERSION) < MAX_VERSION):
        raise RuntimeError(
            f"This version of sksundae requires SUNDIALS >= v{MIN_VERSION} and"
            f" < v{MAX_VERSION}, but found {SUNDIALS_VERSION=}. Please install"
            " a supported version and rebuild sksundae."
        )

    if config.get('SUNDIALS_SINGLE_PRECISION'):
        precision = 'float'
        np_precision = 'np.float32_t'
    elif config.get('SUNDIALS_DOUBLE_PRECISION'):
        precision = 'double'
        np_precision = 'np.float64_t'
    elif config.get('SUNDIALS_EXTENDED_PRECISION'):
        precision = 'long double'
        np_precision = 'np.float128_t'
    else:
        warn("Couldn't find SUNDIALS_PRECISION. Defaulting to double.")
        precision = 'double'
        np_precision = 'np.float64_t'

    if config.get('SUNDIALS_INT32_T'):
        indexsize = 'int'
        np_indexsize = 'np.int32_t'
    elif config.get('SUNDIALS_INT64_T'):
        indexsize = 'long int'
        np_indexsize = 'np.int64_t'
    else:
        warn("Couldn't find SUNDIALS_INDEX_SIZE. Defaulting to int.")
        indexsize = 'int'
        np_indexsize = 'np.int32_t'

    if 'SUNDIALS_SUPERLUMT_ENABLED' in config:
        has_superlu = True
        superlu_threads = config['SUNDIALS_SUPERLUMT_THREAD_TYPE']
    else:
        has_superlu = False
        superlu_threads = None

    superlu_dir = os.path.join(BASE, 'include', 'superlu_mt')
    if has_superlu and not os.path.exists(superlu_dir):
        warn("SuperLU_MT was enabled, but is not on the expected path. It is"
             " expected that the header and library files for SuperLU_MT are"
             " in the same include and lib dirs as SUNDIALS. scikit-SUNDAE"
             " will build without SuperLU_MT. If you need these capabilities"
             " please move your installation and try again.")

        has_superlu = False
        superlu_threads = None

    if 'SUNDIALS_BLAS_LAPACK_ENABLED' in config:
        has_lapack = True
    else:
        has_lapack = False

    with open('src/sksundae/py_config.pxi', 'w') as f:  # Python config
        f.write(f"SUNDIALS_VERSION = \"{SUNDIALS_VERSION}\"\n")
        f.write(f"SUNDIALS_FLOAT_TYPE = \"{precision}\"\n")
        f.write(f"SUNDIALS_INT_TYPE = \"{indexsize}\"\n")
        f.write(f"SUNDIALS_SUPERLUMT_ENABLED = \"{has_superlu}\"\n")
        f.write(f"SUNDIALS_SUPERLUMT_THREAD_TYPE = \"{superlu_threads}\"\n")
        f.write(f"SUNDIALS_BLAS_LAPACK_ENABLED = \"{has_lapack}\"\n")

    with open('src/sksundae/c_config.pxi', 'w') as f:  # C config
        f.write("cimport numpy as np\n\n")

        f.write(f"ctypedef {precision} sunrealtype\n")
        f.write(f"ctypedef {np_precision} DTYPE_t\n\n")

        f.write(f"ctypedef {indexsize} sunindextype\n")
        f.write(f"ctypedef {np_indexsize} INT_TYPE_t\n\n")

    # Specify include_dirs, library_dirs, and libraries for each extension
    SUNDIALS_INCLUDE_DIRS = [np.get_include(), os.path.join(BASE, 'include')]
    SUNDIALS_LIBRARY_DIRS = [os.path.join(BASE, 'lib')]

    LIBRARIES = [
        'sundials_core',
        'sundials_nvecserial',
        'sundials_sunlinsoldense',
        'sundials_sunlinsolband',
        'sundials_sunmatrixdense',
        'sundials_sunmatrixband',
        'sundials_sunmatrixsparse',
        'sundials_sunlinsolspgmr',
        'sundials_sunlinsolspbcgs',
        'sundials_sunlinsolsptfqmr',
    ]

    MACROS = [('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]

    # Optional solvers - SuperLU_MT
    if has_superlu:

        SUNDIALS_INCLUDE_DIRS.append(
            os.path.join(BASE, 'include', 'superlu_mt'),
        )

        LIBRARIES.extend([
            'sundials_sunlinsolsuperlumt',
            'superlu_mt_' + superlu_threads,
        ])

        MACROS.append(('__' + superlu_threads, None))
        MACROS.append(('SUNDIALS_HAS_SUPERLUMT', None))

    # Optional solvers - LAPACK
    if has_lapack:

        LIBRARIES.extend([
            'sundials_sunlinsollapackdense',
            'sundials_sunlinsollapackband',
        ])

        MACROS.append(('SUNDIALS_HAS_LAPACK', None))

    # Define the extension modules
    extensions = [
        setuptools.Extension(
            name='sksundae._cy_common',
            sources=['src/sksundae/_cy_common.pyx'],
            include_dirs=SUNDIALS_INCLUDE_DIRS,
            library_dirs=SUNDIALS_LIBRARY_DIRS,
            libraries=LIBRARIES,
            define_macros=MACROS,
        ),
        setuptools.Extension(
            name='sksundae._cy_cvode',
            sources=['src/sksundae/_cy_cvode.pyx'],
            include_dirs=SUNDIALS_INCLUDE_DIRS,
            library_dirs=SUNDIALS_LIBRARY_DIRS,
            libraries=LIBRARIES + ['sundials_cvode'],
            define_macros=MACROS,
        ),
        setuptools.Extension(
            name='sksundae._cy_ida',
            sources=['src/sksundae/_cy_ida.pyx'],
            include_dirs=SUNDIALS_INCLUDE_DIRS,
            library_dirs=SUNDIALS_LIBRARY_DIRS,
            libraries=LIBRARIES + ['sundials_ida'],
            define_macros=MACROS,
        ),
    ]

    ext_modules = cythonize(
        extensions,
        compiler_directives={'language_level': 3},
    )

    return ext_modules


# Run the setup
BUILD_SDIST = os.environ.get('BUILD_SDIST', 0)

if int(BUILD_SDIST):  # Don't compile extensions if just building sdist
    setuptools.setup(
        include_package_data=True,
    )
else:
    setuptools.setup(
        include_package_data=True,
        ext_modules=get_extensions(),
    )



# _cy_common.pyx

# Dependencies
cimport numpy as np

# Extern cdef headers
from .c_sundials cimport *  # Access to C types
from .c_nvector cimport *  # Access to N_Vector functions
from .c_sunmatrix cimport *  # Access to SUNMatrix functions

# Define float and int types:
# py_config.pxi is created in setup.py. While building the python package, the 
# sundials_config.h header is parsed to determine what precision was used to
# compile the SUNDIALS that is being built against. The settings are saved in
# the pxi file and used here.
include "py_config.pxi"

config = {
    "SUNDIALS_VERSION": SUNDIALS_VERSION,
    "SUNDIALS_FLOAT_TYPE": SUNDIALS_FLOAT_TYPE,
    "SUNDIALS_INT_TYPE": SUNDIALS_INT_TYPE,
    "SUNDIALS_SUPERLUMT_ENABLED": SUNDIALS_SUPERLUMT_ENABLED,
    "SUNDIALS_SUPERLUMT_THREAD_TYPE": SUNDIALS_SUPERLUMT_THREAD_TYPE,
    "SUNDIALS_BLAS_LAPACK_ENABLED": SUNDIALS_BLAS_LAPACK_ENABLED,
}

if SUNDIALS_FLOAT_TYPE == "float":
    from numpy import float32 as DTYPE
elif SUNDIALS_FLOAT_TYPE == "double":
    from numpy import float64 as DTYPE
elif SUNDIALS_FLOAT_TYPE == "long double":
    from numpy import longdouble as DTYPE

if SUNDIALS_INT_TYPE == "int":
    from numpy import int32 as INT_TYPE
elif SUNDIALS_INT_TYPE == "long int":
    from numpy import int64 as INT_TYPE


cdef svec2np(N_Vector nvec, np.ndarray[DTYPE_t, ndim=1] np_array):
    """Fill a numpy array with values from an N_Vector."""
    cdef sunrealtype* nvec_ptr

    nv_ptr = N_VGetArrayPointer(nvec)
    ptr2np(nv_ptr, np_array)


cdef np2svec(np.ndarray[DTYPE_t, ndim=1] np_array, N_Vector nvec):
    """Fill an N_Vector with values from a numpy array."""
    cdef sunrealtype* nv_ptr

    nv_ptr = N_VGetArrayPointer(nvec)
    np2ptr(np_array, nv_ptr)


cdef ptr2np(sunrealtype* nv_ptr, np.ndarray[DTYPE_t, ndim=1] np_array):
    """Fill a numpy array with values from an N_Vector pointer."""
    cdef sunindextype size = <sunindextype> np_array.size

    np_array[:] = <sunrealtype[:size]> nv_ptr


cdef np2ptr(np.ndarray[DTYPE_t, ndim=1] np_array, sunrealtype* nv_ptr):
    """Fill an N_Vector pointer with values from a numpy array."""
    cdef sunindextype size = <sunindextype> np_array.size

    nv_ptr[0:size] = &np_array[0]


cdef np2smat_dense(np.ndarray[DTYPE_t, ndim=2] np_A, SUNMatrix smat):
    """Fill a SUNDenseMatrix with values from a 2D numpy array."""
    cdef sunindextype i, j
    cdef sunindextype M = <sunindextype> np_A.shape[0]
    cdef sunindextype N = <sunindextype> np_A.shape[1]
    cdef sunrealtype** sm_cols = SUNDenseMatrix_Cols(smat)

    for j in range(N):
        for i in range(M):
            sm_cols[j][i] = np_A[i,j]


cdef np2smat_band(np.ndarray[DTYPE_t, ndim=2] np_A, SUNMatrix smat):
    """Fill a SUNBandMatrix with values from a 2D numpy array."""
    cdef sunindextype i, j
    cdef sunindextype N = <sunindextype> np_A.shape[1]
    cdef sunrealtype** sm_cols = SUNBandMatrix_Cols(smat)
    cdef sunindextype lband = SUNBandMatrix_LowerBandwidth(smat)
    cdef sunindextype uband = SUNBandMatrix_UpperBandwidth(smat)
    cdef sunindextype smu = SUNBandMatrix_StoredUpperBandwidth(smat)

    # Indexing is more complex in a SUNBandMatrix, see documentation:
    # https://sundials.readthedocs.io/en/latest/sunmatrix/SUNMatrix_links.html
    for j in range(N):
        i_min = max(0, j - uband)
        i_max = min(N, j + lband + 1)
        for i in range(i_min, i_max):
            sm_cols[j][i-j+smu] = np_A[i,j]


cdef np2smat_sparse1D(np.ndarray[DTYPE_t, ndim=1] np_A, SUNMatrix smat,
                      object sparsity):
    """Fill a SUNSparseMatrix with values from a 1D numpy array."""
    cdef sunindextype nnz, nidx, nptr
    cdef sunrealtype* data = SUNSparseMatrix_Data(smat)
    cdef sunindextype* indices = SUNSparseMatrix_IndexValues(smat)
    cdef sunindextype* indptrs = SUNSparseMatrix_IndexPointers(smat)
    cdef np.ndarray[INT_TYPE_t, ndim=1] np_indices = sparsity.indices
    cdef np.ndarray[INT_TYPE_t, ndim=1] np_indptr = sparsity.indptr

    nnz = <sunindextype> sparsity.nnz
    nidx = <sunindextype> sparsity.indices.size
    nptr = <sunindextype> sparsity.indptr.size

    data[0:nnz] = &np_A[0]
    indices[0:nidx] = <sunindextype*> &np_indices[0]
    indptrs[0:nptr] = <sunindextype*> &np_indptr[0]


cdef np2smat_sparse2D(np.ndarray[DTYPE_t, ndim=2] np_A, SUNMatrix smat,
                      object sparsity):
    """Fill a SUNSparseMatrix with values from a 2D numpy array."""
    cdef sunindextype i, j, start, end, idx
    cdef sunindextype nnz, nidx, nptr, ncols
    cdef sunrealtype* data = SUNSparseMatrix_Data(smat)
    cdef sunindextype* indices = SUNSparseMatrix_IndexValues(smat)
    cdef sunindextype* indptrs = SUNSparseMatrix_IndexPointers(smat)
    cdef np.ndarray[INT_TYPE_t, ndim=1] np_indices = sparsity.indices
    cdef np.ndarray[INT_TYPE_t, ndim=1] np_indptr = sparsity.indptr

    nidx = <sunindextype> sparsity.indices.size
    nptr = <sunindextype> sparsity.indptr.size

    indices[0:nidx] = <sunindextype*> &np_indices[0]
    indptrs[0:nptr] = <sunindextype*> &np_indptr[0]

    ncols = <sunindextype> np_A.shape[1]
    for j in range(ncols):
        start = sparsity.indptr[j]
        end = sparsity.indptr[j+1]

        for i, idx in enumerate(sparsity.indices[start:end]):
            data[start + i] = np_A[idx,j]


cdef np2smat(np.ndarray np_A, SUNMatrix smat, object sparsity):
    """Fill a SUNMatrix with values from np_A using the correct cdef."""
    cdef SUNMatrix_ID matrix_id = SUNMatGetID(smat)

    if matrix_id == SUNMATRIX_DENSE:
        np2smat_dense(np_A, smat)
    elif matrix_id == SUNMATRIX_BAND:
        np2smat_band(np_A, smat)
    elif matrix_id == SUNMATRIX_SPARSE and np_A.ndim == 1:
        np2smat_sparse1D(np_A, smat, sparsity)
    elif matrix_id == SUNMATRIX_SPARSE and np_A.ndim == 2:
        np2smat_sparse2D(np_A, smat, sparsity)
    else:
        raise TypeError("Only 'dense', 'band', or 'sparse' SUNMatrix are"
                        " supported for 'smat'.")



# c_sunmatrix.pxd

from .c_sundials cimport *  # Access to types

# sunmatrix_dense.h
cdef extern from "sunmatrix/sunmatrix_dense.h":
    SUNMatrix SUNDenseMatrix(sunindextype M, sunindextype N, SUNContext ctx)

    sunrealtype** SUNDenseMatrix_Cols(SUNMatrix A)

# sunmatrix_band.h
cdef extern from "sunmatrix/sunmatrix_band.h":
    SUNMatrix SUNBandMatrix(sunindextype M, sunindextype mu, sunindextype ml,
                            SUNContext ctx)

    sunindextype SUNBandMatrix_StoredUpperBandwidth(SUNMatrix A)
    sunindextype SUNBandMatrix_LowerBandwidth(SUNMatrix A)
    sunindextype SUNBandMatrix_UpperBandwidth(SUNMatrix A)
    sunrealtype** SUNBandMatrix_Cols(SUNMatrix A)

# sunmatrix_sparse.h
cdef extern from "sunmatrix/sunmatrix_sparse.h":
    int CSC_MAT
    int CSR_MAT

    SUNMatrix SUNSparseMatrix(sunindextype M, sunindextype N, sunindextype NNZ,
                              int sparsetype, SUNContext ctx)

    sunrealtype* SUNSparseMatrix_Data(SUNMatrix A)
    sunindextype* SUNSparseMatrix_IndexValues(SUNMatrix A)
    sunindextype* SUNSparseMatrix_IndexPointers(SUNMatrix A)


# _cy_cvode.pyx

# Enable embedded signatures for the entire module
# cython: embedsignature=True, embeddedsignature.format='python'

# Standard library
import os
import inspect

from warnings import warn
from numbers import Integral, Real
from typing import Callable, Iterable

# Dependencies
import numpy as np
cimport numpy as np

from scipy import sparse as sp
from scipy.optimize._numdiff import group_columns
from cpython.exc cimport PyErr_CheckSignals, PyErr_Occurred

# Extern cdef headers
from .c_cvode cimport *
from .c_nvector cimport *
from .c_sundials cimport *
from .c_sunmatrix cimport *
from .c_sunlinsol cimport *

# Internal cdef headers
from ._cy_common cimport *
from ._cy_common import DTYPE, INT_TYPE, config  # Python precisions/config

# Local python dependencies
from .utils import RichResult
from .cvode._precond import CVODEPrecond
from .cvode._jactimes import CVODEJacTimes


# Messages shorted from documentation online:
# https://sundials.readthedocs.io/en/latest/cvode/Constants_link.html
CVMESSAGES = {
    0: "Successful function return.",
    1: "Reached specified tstop.",
    2: "Detected one or more events.",
    99: "Succeeded but something unusual happened.",
    -1: "Could not reach endpoint after 'max_num_steps'.",
    -2: "Could not satisfy demanded accuracy for an internal step.",
    -3: "Error tests failed too many times, or reached min step size.",
    -4: "Convergence tests failed too many times, or reached min step size.",
    -5: "Linear solver initialization routine failed.",
    -6: "Linear solver setup function unrecoverably failed.",
    -7: "Linear solver solve function unrecoverably failed.",
    -8: "The right-hand-side function had a non-recoverable error.",
    -9: "The right-hand-side function failed on the first call.",
    -10: "The right-hand-side function had repeated recoverable errors.",
    -11: "'rhsfn' returned recoverable errors, but the solver cannot recover.",
    -12: "Event-detection routine unrecoverably failed.",
    -13: "Nonlinear solver initialization routine failed.",
    -14: "Nonlinear solver setup function failed.",
    -15: "Inequality constraints could not be met.",
    -16: "The nonlinear solver unrecoverably failed.",    
    -20: "A memory allocation request failed.",
    -21: "The integrator's 'mem' argument is NULL.",
    -22: "One of the function inputs is invalid.",
    -23: "Memory was not allocated by a call to CVodeMalloc.",
    -24: "Bad k value. k must be in range 0, 1, ..., order.",
    -25: "Bad t value. t must be within the last step interval.",
    -26: "The output derivate vector is NULL.",
    -27: "The output and initial times are too close to each other.",
    -28: "CVODE experienced a vector operation error.",
    -29: "The projection memory was NULL.",
    -30: "The projection function unrecoverably failed.",
    -31: "The projection function had repeated recoverable errors.",
    -32: "A SUNContext error occurred while initializing the solver.",
    -99: "An unrecognized error occurred within the solver.",
}

LSMESSAGES = {
    0: "Successful function return.",
    -1: "The integrator's 'mem' argument is NULL.",
    -2: "The linear solver has not been initialized.",
    -3: "The linear solver is not compatible with the N_Vector module.",
    -4: "A memory allocation request failed.",
    -5: "The preconditioner module has not been initialized.",
    -6: "The Jacobian function unrecoverably failed.",
    -7: "The Jacobian function had a recoverable error.",
    -8: "An error occurred with the current SUNMatrix module.",
    -9: "An error occurred with the current SUNLinearSolver module.",
}


cdef int _rhsfn_wrapper(sunrealtype t, N_Vector yy, N_Vector yp,
                        void* data) except? -1:
    """Wraps 'rhsfn' by converting between N_Vector and ndarray types."""

    aux = <AuxData> data

    svec2np(yy, aux.np_yy)

    if aux.with_userdata:
        _ = aux.rhsfn(t, aux.np_yy, aux.np_yp, aux.userdata)
    else:
        _ = aux.rhsfn(t, aux.np_yy, aux.np_yp)

    np2svec(aux.np_yp, yp)
    
    return 0


cdef int _eventsfn_wrapper(sunrealtype t, N_Vector yy, sunrealtype* ee,
                           void* data) except? -1:
    """Wraps 'eventsfn' by converting between N_Vector and ndarray types."""

    aux = <AuxData> data

    svec2np(yy, aux.np_yy)

    if aux.with_userdata:
        _ = aux.eventsfn(t, aux.np_yy, aux.np_ee, aux.userdata)
    else:
        _ = aux.eventsfn(t, aux.np_yy, aux.np_ee)

    np2ptr(aux.np_ee, ee)
    
    return 0


cdef int _jacfn_wrapper(sunrealtype t, N_Vector yy, N_Vector yp, SUNMatrix JJ,
                        void* data, N_Vector tmp1, N_Vector tmp2,
                        N_Vector tmp3) except? -1:
    """Wraps 'jacfn' by converting between N_Vector and ndarray types."""
    
    aux = <AuxData> data

    svec2np(yy, aux.np_yy)
    svec2np(yp, aux.np_yp)

    if aux.with_userdata:
        _ = aux.jacfn(t, aux.np_yy, aux.np_yp, aux.np_JJ, aux.userdata)
    else:
        _ = aux.jacfn(t, aux.np_yy, aux.np_yp, aux.np_JJ)

    np2smat(aux.np_JJ, JJ, aux.sparsity)

    return 0


cdef int _psetup_wrapper(sunrealtype t, N_Vector yy, N_Vector yp,
                         sunbooleantype jok, sunbooleantype* jcurPtr,
                         sunrealtype gamma, void* data) except? -1:
    """Wraps 'psetup' by converting between N_Vector and ndarray types."""
    
    aux = <AuxData> data
    psetup = aux.precond.setupfn

    svec2np(yy, aux.np_yy)
    svec2np(yp, aux.np_yp)

    jnew = list((jcurPtr[0],))

    if aux.with_userdata:
        _ = psetup(t, aux.np_yy, aux.np_yp, jok, jnew, gamma, aux.userdata)
    else:
        _ = psetup(t, aux.np_yy, aux.np_yp, jok, jnew, gamma)

    jcurPtr[0] = 1 if jnew[0] else 0

    return 0


cdef int _psolve_wrapper(sunrealtype t, N_Vector yy, N_Vector yp, N_Vector rv,
                         N_Vector zv, sunrealtype gamma, sunrealtype delta,
                         int lr, void* data) except? -1:
    """Wraps 'psolve' by converting between N_Vector and ndarray types."""
    
    aux = <AuxData> data
    psolve = aux.precond.solvefn

    svec2np(yy, aux.np_yy)
    svec2np(yp, aux.np_yp)
    svec2np(rv, aux.np_rv)

    if aux.with_userdata:
        _ = psolve(t, aux.np_yy, aux.np_yp, aux.np_rv, aux.np_zv, gamma,
                   delta, lr, aux.userdata)
    else:
        _ = psolve(t, aux.np_yy, aux.np_yp, aux.np_rv, aux.np_zv, gamma,
                   delta, lr)

    np2svec(aux.np_zv, zv)

    return 0


cdef int _jvsetup_wrapper(sunrealtype t, N_Vector yy, N_Vector yp,
                          void* data) except? -1:
    """Wraps 'jvsetup' by converting between N_Vector and ndarray types."""
    
    aux = <AuxData> data
    jvsetup = aux.jactimes.setupfn

    svec2np(yy, aux.np_yy)
    svec2np(yp, aux.np_yp)

    if aux.with_userdata:
        _ = jvsetup(t, aux.np_yy, aux.np_yp, aux.userdata)
    else:
        _ = jvsetup(t, aux.np_yy, aux.np_yp)

    return 0


cdef int _jvsolve_wrapper(N_Vector vv, N_Vector Jv, sunrealtype t, N_Vector yy,
                          N_Vector yp, void* data, N_Vector tmp) except? -1:
    """Wraps 'jvsolve' by converting between N_Vector and ndarray types."""
    
    aux = <AuxData> data
    jvsolve = aux.jactimes.solvefn

    svec2np(yy, aux.np_yy)
    svec2np(yp, aux.np_yp)
    svec2np(vv, aux.np_vv)

    if aux.with_userdata:
        _ = jvsolve(t, aux.np_yy, aux.np_yp, aux.np_vv, aux.np_Jv, aux.userdata)
    else:
        _ = jvsolve(t, aux.np_yy, aux.np_yp, aux.np_vv, aux.np_Jv)

    np2svec(aux.np_Jv, Jv)

    return 0


cdef void _err_handler(int line, const char* func, const char* file,
                       const char* msg, int err_code, void* err_user_data,
                       SUNContext ctx) except *:
    """Custom error handler for shorter messages (no line or file)."""
    
    if not PyErr_Occurred():
        decoded_func = func.decode("utf-8")
        decoded_msg = msg.decode("utf-8").replace(", ,", ",").strip()
        print(f"\n[{decoded_func}, Error: {err_code}] {decoded_msg}\n")


cdef class AuxData:
    """
    Auxiliary data.
    
    Used to pre-allocate and store numpy arrays in memory, and to carry data
    to function wrappers.

    """
    cdef np.ndarray np_yy       # state variables
    cdef np.ndarray np_yp       # yy time derivatives (aka rhs values fy)
    cdef np.ndarray np_ee       # events array
    cdef np.ndarray np_JJ       # Jacobian matrix
    cdef np.ndarray np_rv       # precond rvec
    cdef np.ndarray np_zv       # precond zvec
    cdef np.ndarray np_vv       # jactimes vv
    cdef np.ndarray np_Jv       # jactimes Jv
    cdef np.ndarray np_cc       # constraints (-2, -1, 0, 1, 2)
    cdef bint with_userdata
    cdef bint is_constrained

    cdef object rhsfn           # Callable
    cdef object userdata        # Any
    cdef object eventsfn        # Callable
    cdef object jacfn           # Callable
    cdef object linsolver       # str
    cdef object sparsity        # csc_matrix
    cdef object precond         # CVODEPrecond
    cdef object jactimes        # CVODEJacTimes

    def __cinit__(self, sunindextype NEQ, object options):
        self.np_yy = np.empty(NEQ, DTYPE)
        self.np_yp = np.empty(NEQ, DTYPE)
        
        self.rhsfn = options["rhsfn"]
        self.userdata = options["userdata"]
        self.with_userdata = 1 if self.userdata is not None else 0

        self.eventsfn = options["eventsfn"]
        self.np_ee = np.empty(options["num_events"], DTYPE)

        self.jacfn = options["jacfn"]
        if self.jacfn is not None:
            self.np_JJ = np.zeros((NEQ, NEQ), DTYPE)
        else:
            self.np_JJ = np.empty(0, DTYPE)

        self.linsolver = options["linsolver"]
        self.sparsity = options["sparsity"]

        self.precond = options["precond"]
        if self.precond is not None:
            self.np_rv = np.empty(NEQ, DTYPE)
            self.np_zv = np.empty(NEQ, DTYPE)
        else:
            self.np_rv = np.empty(0, DTYPE)
            self.np_zv = np.empty(0, DTYPE)

        self.jactimes = options["jactimes"]
        if self.jactimes is not None:
            self.np_vv = np.empty(NEQ, DTYPE)
            self.np_Jv = np.empty(NEQ, DTYPE)
        else:
            self.np_vv = np.empty(0, DTYPE)
            self.np_Jv = np.empty(0, DTYPE)

        constraints_idx = options["constraints_idx"]
        constraints_type = options["constraints_type"]
        if constraints_idx is not None:

            self.is_constrained = True
            self.np_cc = np.zeros(NEQ, INT_TYPE)
            for idx, val in zip(constraints_idx, constraints_type):
                self.np_cc[idx] = val

        else:
            self.is_constrained = False
            self.np_cc = np.zeros(0, INT_TYPE)


cdef class _cvLSSparseDQJac:
    """
    Sparse Jacobian approximation.
    
    This routine generates a sparse difference quotient approximation to the
    system Jacobian. If the dense or banded solvers are being used the routine
    fills a 2D array and is therefore less memory efficient, however, this can
    still improve computational times to evaluate the Jacobian. In contrast,
    if the sparse linear solver is used then a 1D array with NNZ elements is
    used to carry around the output.

    """
    cdef void* mem
    cdef AuxData aux

    cdef object groups      # dict[int, np.ndarray[int]]
    cdef object sparsity    # sparse.csc_matrix, shape(NEQ, NEQ)

    def __cinit__(self, AuxData aux, object sparsity):

        grouped_cols = group_columns(sparsity)
        ngroups = np.max(grouped_cols) + 1

        groups = {}
        for i in range(ngroups):
            cols = np.where(grouped_cols == i)[0]           
            groups[i] = np.array(cols, INT_TYPE)       
            
        self.aux = aux
        self.groups = groups
        self.sparsity = sparsity

    def __call__(
        self,
        sunrealtype t,
        np.ndarray[DTYPE_t, ndim=1] y,
        np.ndarray[DTYPE_t, ndim=1] yp,
        np.ndarray JJ,  # support for 1D (sparse) and 2D (dense, band)
        *userdata,
    ):

        cdef sunrealtype uround, srur
        cdef sunindextype j, k, start, end
        cdef np.ndarray[INT_TYPE_t, ndim=1] cols, indices
        cdef np.ndarray[DTYPE_t, ndim=1] diff, inc, inc_inv, ytemp, yptemp
        
        aux = <AuxData> self.aux
        sparsity = self.sparsity

        ytemp = y.copy()
        yptemp = yp.copy()

        uround = np.finfo(DTYPE).eps
        srur = np.sqrt(uround)

        sign = (y >= 0).astype(float) * 2 - 1
        inc = srur * sign * np.maximum(srur, np.abs(y))

        if aux.is_constrained:
            conj = aux.np_cc

            mask1 = np.abs(conj) == 1
            flip1 = ((y + inc) * conj < 0)
            inc[mask1 & flip1] *= -1

            mask2 = np.abs(conj) == 2
            flip2 = ((y + inc) * conj <= 0)
            inc[mask2 & flip2] *= -1

        inc_inv = 1. / inc

        ngroups = len(self.groups)
        for k in range(ngroups):
            cols = self.groups[k]

            ytemp[cols] += inc[cols]
          
            if aux.with_userdata:
                _ = aux.rhsfn(t, ytemp, yptemp, aux.userdata)
            else:
                _ = aux.rhsfn(t, ytemp, yptemp)

            diff = yptemp - yp
            
            for j in cols:
                start = sparsity.indptr[j]
                end = sparsity.indptr[j+1]

                indices = sparsity.indices[start:end]
                if JJ.ndim == 1:
                    JJ[start:end] = inc_inv[j]*diff[indices]
                elif JJ.ndim == 2:
                    JJ[indices, j] = inc_inv[j]*diff[indices]
                
            ytemp[cols] = y[cols]

    cdef _setup_memory(self, void* mem, sunindextype NEQ):
        """
        Store mem for access to current step, and prep either 1D or 2D array
        for Jacobian storage. 
        
        """
        self.mem = mem
        self.aux.jacfn = self

        if self.aux.linsolver == "sparse":
            nnz = self.sparsity.nnz
            self.aux.np_JJ = np.zeros(nnz, DTYPE)
        else:
            self.aux.np_JJ = np.zeros((NEQ, NEQ), DTYPE)


class CVODEResult(RichResult):
    _order_keys = ["message", "success", "status", "t", "y", "i_events",
                   "t_events", "y_events", "nfev", "njev",]


cdef class CVODE:
    cdef void* mem
    cdef SUNContext ctx
    cdef N_Vector atol
    cdef N_Vector constraints
    cdef N_Vector yy
    cdef SUNMatrix A 
    cdef SUNLinearSolver LS
    cdef sunindextype NEQ
    cdef AuxData aux

    cdef object _size           # int
    cdef object _malloc         # bool - flag for memory allocation
    cdef object _options        # dict[str, Any]
    cdef object _initialized    # bool - flag for init_step completion

    def __cinit__(self, object rhsfn, **options):
        self._free_memory()
        
        self._options = {
            "rhsfn": rhsfn,
            "userdata": None,
            "method": "BDF",
            "first_step": 0.,
            "min_step": 0., 
            "max_step": 0.,
            "rtol": 1e-5, 
            "atol": 1e-6,
            "linsolver": "dense",
            "lband": None,
            "uband": None,
            "sparsity": None,
            "nthreads": None,
            "krylov_dim": None,
            "max_order": 5,
            "max_num_steps": 500,
            "max_nonlin_iters": 3,
            "max_conv_fails": 10,
            "constraints_idx": None,
            "constraints_type": None,
            "eventsfn": None,
            "num_events": 0,
            "jacfn": None,
            "precond": None,
            "jactimes": None,
        }

        invalid_keys = set(options.keys()) - set(self._options.keys())
        if invalid_keys:
            raise ValueError(f"Invalid keyword arguments: {invalid_keys}.")
        
        self._options.update(options)

        method = options.get("method", "").lower()
        if method == "adams" and "max_order" not in options:
            self._options["max_order"] = 12

        _check_options(self._options)

        self._initialized = False

    cdef _create_linsolver(self):
        iterative = {"gmres", "bicgstab", "tfqmr"}
        direct = {"dense", "lapackdense", "band", "lapackband", "sparse"}

        linsolver = self._options["linsolver"].lower()

        if "band" in linsolver:
            uband = <int> self._options["uband"]
            lband = <int> self._options["lband"]
        elif linsolver in iterative:
            maxl = <int> self._options["krylov_dim"]

            precond = self._options["precond"]
            prectype = SUN_PREC_NONE if precond is None else precond._prectype
        
        if linsolver == "dense":
            self.A = SUNDenseMatrix(self.NEQ, self.NEQ, self.ctx)
            self.LS = SUNLinSol_Dense(self.yy, self.A, self.ctx)

        elif linsolver == "lapackdense":
            self.A = SUNDenseMatrix(self.NEQ, self.NEQ, self.ctx)
            self.LS = SUNLinSol_LapackDense(self.yy, self.A, self.ctx)

        elif linsolver == "band":
            self.A = SUNBandMatrix(self.NEQ, uband, lband, self.ctx)
            self.LS = SUNLinSol_Band(self.yy, self.A, self.ctx)

        elif linsolver == "lapackband":
            self.A = SUNBandMatrix(self.NEQ, uband, lband, self.ctx)
            self.LS = SUNLinSol_LapackBand(self.yy, self.A, self.ctx)

        elif linsolver == "sparse":
            nnz = <sunindextype> self._options["sparsity"].nnz
            nthreads = <int> self._options["nthreads"]

            self.A = SUNSparseMatrix(self.NEQ, self.NEQ, nnz, CSC_MAT, self.ctx)
            self.LS = SUNLinSol_SuperLUMT(self.yy, self.A, nthreads, self.ctx)

        elif linsolver == "gmres":
            self.LS = SUNLinSol_SPGMR(self.yy, prectype, maxl, self.ctx)

        elif linsolver == "bicgstab":
            self.LS = SUNLinSol_SPBCGS(self.yy, prectype, maxl, self.ctx)

        elif linsolver == "tfqmr":
            self.LS = SUNLinSol_SPTFQMR(self.yy, prectype, maxl, self.ctx)

        if (linsolver in direct) and (self.A is NULL):
            raise MemoryError("SUNMatrix constructor returned NULL.")
        elif self.LS is NULL:
            raise MemoryError("SUNLinSol constructor returned NULL.")

    cdef _set_tolerances(self):
        rtol = self._options["rtol"]
        atol = self._options["atol"]

        if isinstance(atol, Iterable):
            rtol = <sunrealtype> rtol
            atol = np.asarray(atol, DTYPE)

            if len(atol) != self.NEQ:
                raise ValueError(f"'atol' length ({atol.size}) differs from"
                                 f" problem size ({self.NEQ}).")

            self.atol = N_VNew_Serial(atol.size, self.ctx)
            np2svec(atol, self.atol)

            flag = CVodeSVtolerances(self.mem, rtol, self.atol)

        else:
            rtol = <sunrealtype> rtol
            atol = <sunrealtype> atol 

            flag = CVodeSStolerances(self.mem, rtol, atol)

        if flag < 0:
            raise RuntimeError("CVodetolerances - " + CVMESSAGES[flag])

    cdef _free_memory(self):
        if self.mem is not NULL:
            CVodeFree(&self.mem)
            self.mem = NULL

        if self.ctx is not NULL:
            SUNContext_Free(&self.ctx)
            self.ctx = NULL

        if self.atol is not NULL:
            N_VDestroy(self.atol)
            self.atol = NULL

        if self.constraints is not NULL:
            N_VDestroy(self.constraints)
            self.constraints = NULL

        if self.yy is not NULL:
            N_VDestroy(self.yy)
            self.yy = NULL

        if self.A is not NULL:
            SUNMatDestroy(self.A)
            self.A = NULL

        if self.LS is not NULL:
            SUNLinSolFree(self.LS)
            self.LS = NULL

        self._size = None
        self._malloc = False

    cdef _setup(self, sunrealtype t0, np.ndarray[DTYPE_t, ndim=1] y0):

        # Enumerated steps roughly correspond to the SUNDIALS documentation,
        # available at https://sundials.readthedocs.io/en/latest/cvode/Usage.

        cdef int flag
        cdef np.ndarray np_eventsdir

        # 1) Initialize parallel environment (skip, only use serial here)

        # 2) Create sundials context object
        flag = SUNContext_Create(SUN_COMM_NULL, &self.ctx)
        if flag < 0:
            raise RuntimeError(f"SUNContext_Create failed with {flag=}.")

        # 3) Set problem dimensions
        
        # 4) Create vectors of initial values        
        self.NEQ = <sunindextype> y0.size
        self.aux = AuxData(self.NEQ, self._options)

        self.yy = N_VNew_Serial(self.NEQ, self.ctx)
        if self.yy is NULL:
            raise MemoryError("N_VNew_Serial returned a NULL pointer for yy.")
        
        np2svec(y0.copy(), self.yy)

        # 5) Create CVODE object
        if self._options["method"].lower() == "adams":
            method = CV_ADAMS
        elif self._options["method"].lower() == "bdf":
            method = CV_BDF

        self.mem = CVodeCreate(method, self.ctx)
        if self.mem is NULL:
            raise MemoryError("CVodeCreate returned a NULL pointer for 'mem'.")

        # Attach AuxData - usually done in step 16, but needs to occur here,
        # before attaching preconditioner.
        flag = CVodeSetUserData(self.mem, <void*> self.aux)
        if flag < 0:
            raise RuntimeError("CVodeSetUserData - " + CVMESSAGES[flag])

        # 6) Initialize CVODE solver
        flag = CVodeInit(self.mem, _rhsfn_wrapper, t0, self.yy)
        if flag < 0:
            raise RuntimeError("CVodeInit - " + CVMESSAGES[flag])

        # 7) Specify integration tolerances
        self._set_tolerances()

        # 8) and 9) Create matrix and linear solver - they must match
        self._create_linsolver()

        # 10) Attach the linear solver
        flag = CVodeSetLinearSolver(self.mem, self.LS, self.A)
        if flag < 0:
            raise RuntimeError("CVodeSetLinearSolver - " + LSMESSAGES[flag])

        # 11) Set linear solver optional inputs
        sparsity = self._options["sparsity"]
        if sparsity is not None:
            spjac = _cvLSSparseDQJac(self.aux, sparsity)
            spjac._setup_memory(self.mem, self.NEQ)
            
            if self._options["jacfn"] is None:
                self._options["jacfn"] = spjac 

        jacfn = self._options["jacfn"]
        if jacfn:
            flag = CVodeSetJacFn(self.mem, _jacfn_wrapper)
            if flag < 0:
                raise RuntimeError("CVodeSetJacFn - " + LSMESSAGES[flag])

        precond = self._options["precond"]
        if precond is None:
            pass
        elif precond.setupfn is None:
            flag = CVodeSetPreconditioner(self.mem, NULL, _psolve_wrapper)
            if flag < 0:
                raise RuntimeError("CVodeSetPrecond - " + LSMESSAGES[flag])
        else:
            flag = CVodeSetPreconditioner(self.mem, _psetup_wrapper,
                                          _psolve_wrapper)
            if flag < 0:
                raise RuntimeError("CVodeSetPrecond - " + LSMESSAGES[flag])

        jactimes = self._options["jactimes"]
        if jactimes is None:
            pass
        elif jactimes.setupfn is None:
            flag = CVodeSetJacTimes(self.mem, NULL, _jvsolve_wrapper)
            if flag < 0:
                raise RuntimeError("CVodeSetJacTimes - " + LSMESSAGES[flag])
        else:
            flag = CVodeSetJacTimes(self.mem, _jvsetup_wrapper,
                                    _jvsolve_wrapper)
            if flag < 0:
                raise RuntimeError("CVodeSetJacTimes - " + LSMESSAGES[flag])

        # 12) Create nonlinear solver object (skip, use default Newton solver)

        # 13) Attach nonlinear solver module (skip, use default Newton solver)

        # 14) Set nonlinear solver optional inputs
        cdef int max_nonlin_iters = <int> self._options["max_nonlin_iters"]
        flag = CVodeSetMaxNonlinIters(self.mem, max_nonlin_iters)
        if flag < 0:
            raise RuntimeError("CVodeSetMaxNonlinIters - " + CVMESSAGES[flag])

        cdef int max_conv_fails = <int> self._options["max_conv_fails"]
        flag = CVodeSetMaxConvFails(self.mem, max_conv_fails)
        if flag < 0:
            raise RuntimeError("CVodeSetMaxConvFails - " + CVMESSAGES[flag])

        # 15) Specify rootfinding problem
        eventsfn = self._options["eventsfn"]
        num_events = self._options["num_events"]
        if eventsfn:
            flag = CVodeRootInit(self.mem, <int> num_events, _eventsfn_wrapper)
            if flag < 0:
                raise RuntimeError("CVodeRootInit - " + CVMESSAGES[flag])

            np_eventsdir = np.array(eventsfn.direction, INT_TYPE)

            flag = CVodeSetRootDirection(self.mem, <int*> np_eventsdir.data)
            if flag < 0:
                raise RuntimeError("CVSetRootDirection - " + CVMESSAGES[flag])

        # 16) Set optional inputs
        SUNContext_ClearErrHandlers(self.ctx)
        SUNContext_PushErrHandler(self.ctx, _err_handler, NULL)

        cdef sunrealtype first_step = <sunrealtype> self._options["first_step"] 
        flag = CVodeSetInitStep(self.mem, first_step)
        if flag < 0:
            raise RuntimeError("CVodeSetInitStep - " + CVMESSAGES[flag])

        cdef sunrealtype min_step = <sunrealtype> self._options["min_step"] 
        flag = CVodeSetMinStep(self.mem, min_step)
        if flag < 0:
            raise RuntimeError("CVodeSetMinStep - " + CVMESSAGES[flag])

        cdef sunrealtype max_step = <sunrealtype> self._options["max_step"] 
        flag = CVodeSetMaxStep(self.mem, max_step)
        if flag < 0:
            raise RuntimeError("CVodeSetMaxStep - " + CVMESSAGES[flag])

        cdef int max_order = <int> self._options["max_order"]
        flag = CVodeSetMaxOrd(self.mem, max_order)
        if flag < 0:
            raise RuntimeError("CVodeSetMaxOrd - " + CVMESSAGES[flag])

        cdef long int max_num_steps = <long int> self._options["max_num_steps"]
        flag = CVodeSetMaxNumSteps(self.mem, max_num_steps)
        if flag < 0:
            raise RuntimeError("CVodeSetMaxNumSteps - " + CVMESSAGES[flag])

        constraints_idx = self._options["constraints_idx"]
        constraints_type = self._options["constraints_type"]
        if constraints_idx is not None:

            np_constraints = np.zeros(self.NEQ, DTYPE)
            for idx, val in zip(constraints_idx, constraints_type):
                np_constraints[idx] = val

            self.constraints = N_VNew_Serial(self.NEQ, self.ctx)
            np2svec(np_constraints, self.constraints)

            flag = CVodeSetConstraints(self.mem, self.constraints)
            if flag < 0:
                raise RuntimeError("CVodeSetConstraints - " + CVMESSAGES[flag])

        self._size = self.NEQ
        self._malloc = True
        
        return flag

    cdef _init_step(self, sunrealtype t0, np.ndarray[DTYPE_t, ndim=1] y0):
        cdef int flag

        yy_tmp = y0.copy()

        # Memory allocation and settings steps handled in _setup()... only runs
        # on first call, or if the size of the system changes.

        if not self._malloc:
            flag = self._setup(t0, y0)

        elif self._size != y0.size:
            self._free_memory()
            flag = self._setup(t0, y0)

        else:
            np2svec(yy_tmp, self.yy)
            
            flag = CVodeReInit(self.mem, t0, self.yy)
            if flag < 0:
                raise RuntimeError("CVodeReInit - " + CVMESSAGES[flag])

        self._initialized = True

        # Construct result instance to return
        svec2np(self.yy, yy_tmp)

        nfev, njev = _collect_stats(self.mem)

        result = CVODEResult(
            message=CVMESSAGES[flag], success=flag >= 0, status=flag,
            t=t0, y=yy_tmp.copy(), i_events=None, t_events=None, y_events=None,
            nfev=nfev, njev=njev,
        )

        return result

    cdef _step(self, sunrealtype tt, object method, object tstop):
        cdef int itask
        cdef sunrealtype tout

        # Setup step type:
        if method == "normal":  # output solution at tt
            itask = CV_NORMAL
        elif method == "onestep":  # output after one internal step toward tt
            itask = CV_ONE_STEP

        if isinstance(tstop, Real):
            flag = CVodeSetStopTime(self.mem, <sunrealtype> tstop)
            if flag < 0:
                raise RuntimeError("CVodeSetStopTime - " + CVMESSAGES[flag])

        yy_tmp = self.aux.np_yy
        
        # 17) Advance solution in time
        flag = CVode(self.mem, tt, self.yy, &tout, itask)

        svec2np(self.yy, yy_tmp)

        if flag == CV_ROOT_RETURN:
            _ = _handle_events(self.mem, self.aux, tout, yy_tmp)

        if self.aux.eventsfn:
            i_ev, t_ev, y_ev = _collect_events(self.aux)
        else:
            i_ev, t_ev, y_ev = [None]*3

        nfev, njev = _collect_stats(self.mem)

        result = CVODEResult(
            message=CVMESSAGES[flag], success=flag >= 0, status=flag,
            t=tout, y=yy_tmp.copy(), i_events=i_ev, t_events=t_ev,
            y_events=y_ev, nfev=nfev, njev=njev,
        )

        flag = CVodeClearStopTime(self.mem)
        if flag < 0:
            raise RuntimeError("CVodeClearStopTime - " + CVMESSAGES[flag])

        return result

    cdef _normal_solve(self, np.ndarray[DTYPE_t, ndim=1] tspan,
                             np.ndarray[DTYPE_t, ndim=1] y0,
        ):

        cdef int ind
        cdef int flag
        cdef int stop
        cdef sunrealtype tt
        cdef sunrealtype tend

        _ = self._init_step(tspan[0], y0)

        # Setup solution storage
        tt_out = np.empty(tspan.size, DTYPE)
        yy_out = np.empty((tspan.size, self.NEQ), DTYPE)

        yy_tmp = self.aux.np_yy

        tt_out[0] = tspan[0]
        svec2np(self.yy, yy_out[0, :])

        # 17) Advance solution in time
        stop = 0
        ind = 1

        flag = CVodeSetStopTime(self.mem, <sunrealtype> tspan[-1])
        if flag < 0:
            raise RuntimeError("CVodeSetStopTime - " + CVMESSAGES[flag])

        while True:
            tend = tspan[ind]

            flag = CVode(self.mem, tend, self.yy, &tt, CV_NORMAL)

            svec2np(self.yy, yy_tmp)

            if flag == CV_ROOT_RETURN:
                stop = _handle_events(self.mem, self.aux, tt, yy_tmp)
            elif flag == CV_TSTOP_RETURN:
                stop = 1
            elif ind == len(tspan) - 1:
                stop = 1
            elif flag < 0:
                stop = 1

            if flag == CV_ROOT_RETURN and not stop:
                pass
            else:
                tt_out[ind] = tt
                yy_out[ind, :] = yy_tmp

                ind += 1

            if stop:
                break
            elif PyErr_CheckSignals() == -1:
                return

        if self.aux.eventsfn:
            i_ev, t_ev, y_ev = _collect_events(self.aux)
        else:
            i_ev, t_ev, y_ev = [None]*3

        nfev, njev = _collect_stats(self.mem)

        result = CVODEResult(
            message=CVMESSAGES[flag], success=flag >= 0, status=flag,
            t=tt_out[:ind], y=yy_out[:ind], i_events=i_ev, t_events=t_ev,
            y_events=y_ev, nfev=nfev, njev=njev,
        )

        flag = CVodeClearStopTime(self.mem)
        if flag < 0:
            raise RuntimeError("CVodeClearStopTime - " + CVMESSAGES[flag])      

        return result

    cdef _onestep_solve(self, np.ndarray[DTYPE_t, ndim=1] tspan,
                              np.ndarray[DTYPE_t, ndim=1] y0,
        ):

        cdef int ind
        cdef int flag
        cdef int stop
        cdef sunrealtype tt
        cdef sunrealtype tend

        _ = self._init_step(tspan[0], y0)

        # Setup solution storage
        # Pre-allocate some memory (for 1000 time steps) to fill. Periodically
        # add 500 more more in if the pre-allocated memory gets filled.
        tt_out = np.empty(1000, DTYPE)
        yy_out = np.empty((1000, self.NEQ), DTYPE)

        extra_t = np.empty(500, DTYPE)
        extra_y = np.empty((500, self.NEQ), DTYPE)

        yy_tmp = self.aux.np_yy

        tt_out[0] = tspan[0]
        svec2np(self.yy, yy_out[0, :])

        tend = tspan[-1]
        stop = 0
        ind = 1

        flag = CVodeSetStopTime(self.mem, tend)
        if flag < 0:
            raise RuntimeError("CVodeSetStopTime - " + CVMESSAGES[flag])

        # 17) Advance solution in time
        while True:
            flag = CVode(self.mem, tend, self.yy, &tt, CV_ONE_STEP)

            svec2np(self.yy, yy_tmp)

            if flag == CV_ROOT_RETURN:
                stop = _handle_events(self.mem, self.aux, tt, yy_tmp)
            elif flag == CV_TSTOP_RETURN:
                stop = 1
            elif flag < 0:
                stop = 1

            if ind == tt_out.size - 1:
                tt_out = np.concatenate((tt_out, extra_t))
                yy_out = np.concatenate((yy_out, extra_y))

            if flag == CV_ROOT_RETURN and not stop:
                pass
            else:
                tt_out[ind] = tt
                yy_out[ind, :] = yy_tmp

                ind += 1

            if stop:
                break
            elif PyErr_CheckSignals() == -1:
                return

        if self.aux.eventsfn:
            i_ev, t_ev, y_ev = _collect_events(self.aux)
        else:
            i_ev, t_ev, y_ev = [None]*3

        nfev, njev = _collect_stats(self.mem)

        result = CVODEResult(
            message=CVMESSAGES[flag], success=flag >= 0, status=flag,
            t=tt_out[:ind], y=yy_out[:ind], i_events=i_ev, t_events=t_ev,
            y_events=y_ev, nfev=nfev, njev=njev,
        )

        flag = CVodeClearStopTime(self.mem)
        if flag < 0:
            raise RuntimeError("CVodeClearStopTime - " + CVMESSAGES[flag])

        return result

    def init_step(self, DTYPE_t t0, object y0):
        
        y0 = np.asarray(y0, DTYPE)
        
        return self._init_step(t0, y0)

    def step(self, DTYPE_t t, object method, object tstop):

        method = method.lower()
        valid = {"normal", "onestep"}
        if method not in valid:
            raise ValueError(f"'method' is invalid. Valid values are {valid}.")
        elif not self._initialized:
            raise ValueError("'init_step' must be run prior to 'step'.")

        if tstop is None:
            pass
        elif not isinstance(tstop, Real):
            raise TypeError("'tstop' must be type float, or None.")
        
        return self._step(t, method, tstop)

    def solve(self, object tspan, object y0):

        tspan = np.asarray(tspan, DTYPE)
        y0 = np.asarray(y0, DTYPE)

        diff = np.diff(tspan)
        if not all(diff > 0) ^ all(diff < 0):
            raise ValueError("'tspan' must stictly increase or decrease.")

        if tspan.size > 2:
            soln = self._normal_solve(tspan, y0)
        elif tspan.size == 2:
            soln = self._onestep_solve(tspan, y0)
        else:
            raise ValueError("'tspan' length must be >= 2.")

        self._initialized = False 

        return soln

    def __dealloc__(self):
        self._free_memory()


cdef _prepare_events(object eventsfn, int num_events):

    # eventsfn.terminal
    if not hasattr(eventsfn, "terminal"):
        eventsfn.terminal = [True]*num_events

    terminal = eventsfn.terminal
    if not isinstance(terminal, Iterable):
        raise TypeError("'eventsfn.terminal' must be type Iterable.")
    elif not all(isinstance(x, (bool, Integral)) for x in terminal):
        raise TypeError("All 'eventsfn.terminal' values must be bool or int.")
    elif not all(int(x) >= 0 for x in terminal):
        raise ValueError("At least one 'eventsfn.terminal' value is invalid."
                         " Values must be interpretable as int(x) >= 0.")
    elif len(terminal) != num_events:
        raise ValueError("'eventsfn.terminal' length != 'num_events'.")

    # eventsfn.direction
    if not hasattr(eventsfn, "direction"):
        eventsfn.direction = [0]*num_events

    direction = eventsfn.direction
    if not isinstance(direction, Iterable):
        raise TypeError("'eventsfn.direction' must be type Iterable.")
    elif not all(x in (-1, 0, 1) for x in direction):
        raise ValueError(f"At least one 'eventsfn.direction' value is invalid."
                          " Values must be in {-1, 0, 1}.")
    elif len(direction) != num_events:
        raise ValueError("'eventsfn.direction' length != 'num_events'.")

    # add extra fields for _handle_events function
    eventsfn._i_tmp = np.zeros(num_events, INT_TYPE)
    eventsfn._i_cnt = np.zeros(num_events, INT_TYPE)

    eventsfn._i = []
    eventsfn._t = []
    eventsfn._y = []

    eventsfn._max_events = []
    for i, term in enumerate(terminal):
        if term == False:
            eventsfn._max_events.append(np.inf)
        elif term == True:
            eventsfn._max_events.append(1)
        else:
            eventsfn._max_events.append(term)   


cdef _handle_events(void* mem, AuxData aux, sunrealtype tt, np.ndarray yy_tmp):

    cdef int flag
    cdef int stop
    cdef np.ndarray i_tmp

    fn = aux.eventsfn
    i_tmp = fn._i_tmp

    flag = CVodeGetRootInfo(mem, <int*> i_tmp.data)
    if flag < 0:
        raise RuntimeError("CVodeGetRootInfo - " + CVMESSAGES[flag])

    fn._i.append(i_tmp.copy())
    fn._t.append(tt)
    fn._y.append(yy_tmp.copy())

    fn._i_cnt[i_tmp != 0] += 1
    if any(fn._i_cnt >= fn._max_events):
        stop = 1
    else:
        stop = 0

    return stop


cdef _collect_events(AuxData aux):

    fn = aux.eventsfn

    i_events = np.asarray(fn._i, INT_TYPE) if fn._i else None
    t_events = np.asarray(fn._t, DTYPE) if fn._t else None
    y_events = np.asarray(fn._y, DTYPE) if fn._y else None

    return i_events, t_events, y_events


cdef _collect_stats(void* mem):
    cdef long int nfev
    cdef long int njev

    flag = CVodeGetNumRhsEvals(mem, &nfev)
    if flag < 0:
        raise RuntimeError("CVodeGetNumRhsEvals - " + CVMESSAGES[flag])

    flag = CVodeGetNumJacEvals(mem, &njev)
    if flag < 0:
        raise RuntimeError("CVodeGetNumJacEvals - " + LSMESSAGES[flag])

    return nfev, njev


def _check_signature(name: str, func: Callable, expected: tuple[int]) -> int:
    """Check 'rhsfn', 'eventsfn', and 'jacfn' signatures."""

    signature = inspect.signature(func)
    parameters = signature.parameters.values()

    has_args = any([p.kind == inspect._VAR_POSITIONAL for p in parameters])
    has_kwargs = any([p.kind == inspect._VAR_KEYWORD for p in parameters])

    if has_args or has_kwargs:
        raise ValueError(f"'{name}' cannot include *args or **kwargs.")

    if name == "resfn" and len(parameters) not in expected:
        raise ValueError(f"'{name}' has an invalid signature. It must only"
                          " have 3 (w/o userdata) or 4 (w/ userdata) args.")
    elif len(parameters) not in expected:
        raise ValueError(f"'{name}' signature is inconsistent with 'rhsfn'."
                         " look for a missing or extraneous 'userdata' arg.")
    
    if name == "rhsfn" and len(parameters) == 3:
        with_userdata = 0
    elif name == "rhsfn" and len(parameters) == 4:
        with_userdata = 1
    else:
        with_userdata = None

    return with_userdata


def _check_options(options: dict) -> None:

    # rhsfn
    if not isinstance(options["rhsfn"], Callable):
        raise TypeError("'rhsfn' must be type Callable.")
    else:
        expected = (3, 4)
        with_userdata = _check_signature("rhsfn", options["rhsfn"], expected)

    # userdata    
    if with_userdata and options["userdata"] is None:
        raise ValueError("'userdata' cannot be None if 'rhsfn' has 4 args.")
    elif options["userdata"] and not with_userdata:
        warn("'userdata' will be ignored since 'rhsfn' only has 3 args.")

    # method
    valid = {"adams", "bdf"}
    method = options["method"].lower()
    if not isinstance(method, str):
        raise TypeError("'method' must be type str.")
    elif method not in valid:
        raise ValueError(f"{method=} is invalid. Must be in {valid}.")
    
    # first_step
    if not isinstance(options["first_step"], Real):
        raise TypeError("'first_step' must be type float.")
    elif options["first_step"] < 0.:
        raise ValueError("'first_step' must be positive or zero.")
        
    # min_step
    if not isinstance(options["min_step"], Real):
        raise TypeError("'min_step' must be type float.")
    elif options["min_step"] < 0.:
        raise ValueError("'min_step' must be positive or zero.")

    # max_step
    if not isinstance(options["max_step"], Real):
        raise TypeError("'max_step' must be type float.")
    elif options["max_step"] < 0.:
        raise ValueError("'max_step' must be positive or zero.")
    elif options["max_step"] < options["min_step"]:
        raise ValueError("'max_step' cannot be smaller than 'min_step'.")

    # rtol
    if not isinstance(options["rtol"], Real):
        raise TypeError("'rtol' must be type float.")

    # atol
    if isinstance(options["atol"], Real):
        pass
    elif not isinstance(options["atol"], Iterable):
        raise TypeError("'atol' must be type float or Iterable[float].")
    elif not all(isinstance(x, Real) for x in options["atol"]):
        raise TypeError("When iterable, all 'atol' values must be float.")

    # linsolver
    iterative = {"gmres", "bicgstab", "tfqmr"}
    direct = {"dense", "lapackdense", "band", "lapackband", "sparse"}
    
    valid = iterative | direct

    linsolver = options["linsolver"].lower()
    if not isinstance(linsolver, str):
        raise TypeError("'linsolver' must be type str.")
    elif linsolver not in valid:
        raise ValueError(f"{linsolver=} is invalid. Must be in {valid}.")

    if "lapack" in linsolver and not config["SUNDIALS_BLAS_LAPACK_ENABLED"]:
        raise ValueError("Cannot use 'lapack*' solvers. LAPACK not enabled.")

    if linsolver == "sparse" and not config["SUNDIALS_SUPERLUMT_ENABLED"]:
        raise ValueError("Cannot use 'sparse' solver. SuperLU_MT not enabled.")

    # lband
    lband = options["lband"]
    if lband is None:
        pass
    elif not isinstance(lband, Integral):
        raise TypeError("'lband' must be type int.")
    elif lband < 0:
        raise ValueError("'lband' must be positive or zero.")

    # uband
    uband = options["uband"]
    if uband is None:
        pass
    elif not isinstance(uband, Integral):
        raise TypeError("'uband' must be type int.")
    elif uband < 0:
        raise ValueError("'uband' must be positive or zero.")

    # sparsity
    sparsity = options["sparsity"]
    if sparsity is None:
        pass 
    elif sp.issparse(sparsity):
        sparsity = sparsity.tocsc()
    elif isinstance(sparsity, np.ndarray):
        sparsity = sp.csc_matrix(sparsity)
    else:
        raise TypeError("'sparsity' must be either a sparse scipy matrix or a"
                        " 2D numpy array.")

    if sparsity is None:
        pass
    elif sparsity.shape[0] != sparsity.shape[1]:
        raise ValueError("'sparsity' must be a square matrix.")

    if (linsolver in iterative) and (sparsity is not None):
        raise ValueError("'sparsity' is not compatitle with iterative linear"
                         f" solvers: {iterative}.")

    options["sparsity"] = sparsity  # save update to CSC sparse, if done

    # nthreads
    ncpu_cores = os.cpu_count()
    nthreads = options["nthreads"]
    if linsolver == "sparse" and sparsity is not None:
        if nthreads is None:
            nthreads = 1
        elif not isinstance(nthreads, Integral):
            raise TypeError("'nthreads' must be type int.")
        elif nthreads == 0:
            nthreads = 1
        elif nthreads <= -1 or nthreads > ncpu_cores:
            nthreads = ncpu_cores

        options["nthreads"] = nthreads  # save defaults update, if done

    # krylov_dim
    krylov_dim = options["krylov_dim"]
    if linsolver in iterative:
        if krylov_dim is None:
            krylov_dim = 5
        elif not isinstance(krylov_dim, Integral):
            raise TypeError("'krylov_dim' must be type int.")
        elif krylov_dim <= 0:
            krylov_dim = 5

        options["krylov_dim"] = krylov_dim  # save defaults update, if done

    elif (linsolver in direct) and (krylov_dim is not None):
        warn("Ignoring 'krylov_dim' since 'linsolver' is not iterative.")

    # consistency between linsolver and lband/uband
    if ("band" in linsolver) and (lband is None or uband is None):
        raise ValueError("banded solver requires integer 'lband', 'uband'.")
    elif ("band" not in linsolver) and (lband is not None or uband is not None):
        warn("Ignoring 'lband', 'uband' since 'linsolver' is not banded.")

    # consistency between linsolver and sparsity/nthreads
    if linsolver == "sparse" and sparsity is None:
        raise ValueError("'sparse' solver requires 'sparsity' not be None.")

    elif linsolver != "sparse" and nthreads is not None:
        warn("Ignoring 'nthreads' since 'linsolver' is not 'sparse'.")

    # max_order
    if method == "bdf":
        max_allowed = 5
    elif method == "adams":
        max_allowed = 12

    if not isinstance(options["max_order"], Integral):
        raise TypeError("'max_order' must be type int.")
    elif options["max_order"] < 1 or options["max_order"] > max_allowed:
        raise ValueError(f"'max_order' must be in range [1, {max_allowed}].")

    # max_num_steps
    if not isinstance(options["max_num_steps"], Integral):
        raise TypeError("'max_num_steps' must be type int.")
    elif not options["max_num_steps"] > 0:
        raise ValueError("'max_num_steps' must be > 0.")

    # max_nonlin_iters
    if not isinstance(options["max_nonlin_iters"], Integral):
        raise TypeError("'max_nonlin_iters' must be type int.")
    elif not options["max_nonlin_iters"] > 0:
        raise ValueError("'max_nonlin_iters' must be > 0.")

    # max_conv_fails
    if not isinstance(options["max_conv_fails"], Integral):
        raise TypeError("'max_conv_fails' must be type int.")
    elif not options["max_conv_fails"] > 0:
        raise ValueError("'max_conv_fails' must be > 0.")

    # constraints_idx
    constraints_idx = options["constraints_idx"]
    if constraints_idx is None:
        pass
    elif not isinstance(constraints_idx, Iterable):
        raise TypeError("'constraints_idx' must be type Iterable.")
    elif not all(isinstance(x, Integral) for x in constraints_idx):
        raise TypeError("All 'constraints_idx' values must be type int.")

    # constraints_type
    constraints_type = options["constraints_type"]
    if constraints_type is None:
        pass
    elif not isinstance(constraints_type, Iterable):
        raise TypeError("'constraints_type' must be type Iterable")
    elif not all(x in (-2, -1, 1, 2) for x in constraints_type):
        raise ValueError(f"At least one 'constraints_type' value is invalid."
                          " Values must be in {-2, -1, 1, 2}.")

    # consistency between constraints index and types
    if constraints_idx is None and constraints_type is None:
        pass 
    elif (constraints_idx is None) ^ (constraints_type is None):
        raise ValueError("'constraints_idx' and 'constraints_type' must both"
                         " be set or both be None.")
    elif len(constraints_idx) != len(constraints_type):
        raise ValueError("'constraints_idx' and 'constraints_type' lengths"
                         " must be the same.")

    # eventsfn
    eventsfn = options["eventsfn"]
    if eventsfn is None:
        pass
    elif not isinstance(eventsfn, Callable):
        raise TypeError("'eventsfn' must be type Callable.")
    else:
        expected = (3 + with_userdata,)
        _ = _check_signature("eventsfn", eventsfn, expected)

    # num_events
    num_events = options["num_events"]    
    if num_events == 0:
        pass
    elif not isinstance(num_events, Integral):
        raise TypeError("'num_events' must be type int.")
    elif num_events < 0:
        raise ValueError("'num_events' must be positive or zero.")

    # consistency between eventsfn and num_events
    if eventsfn and not num_events:
        raise ValueError("'num_events' cannot be 0 if 'eventsfn' is set.")
    elif num_events and not eventsfn:
        warn("'num_events' will be ignored since 'eventsfn' is not set.")

    # prepare events if eventsfn is not None
    if eventsfn:
        _prepare_events(eventsfn, num_events)
        
    # jacfn
    jacfn = options["jacfn"]
    if jacfn is None:
        pass
    elif not isinstance(jacfn, Callable):
        raise TypeError("'jacfn' must be type Callable.")
    else:
        expected = (4 + with_userdata,)
        _ = _check_signature("jacfn", jacfn, expected)

    if jacfn and linsolver in iterative:
        raise ValueError("'jacfn' is not compatitle with iterative linear"
                         f" solvers: {iterative}.")
    
    # preference between sparsity and jacfn
    if (sparsity is not None) and (jacfn is not None):
        warn("Sparse Jacobian approximation will be ignored in favor of"
             " 'jacfn'.")

    # precond
    precond = options["precond"]
    if precond is None:
        pass
    elif not isinstance(precond, CVODEPrecond):
        raise TypeError("'precond' must be type CVODEPrecond.")
    else:
        side = {
            "left": SUN_PREC_LEFT,
            "right": SUN_PREC_RIGHT,
            "both": SUN_PREC_BOTH,
        }
        precond._prectype = side[precond.side]

        if precond.setupfn:
            expected = (6 + with_userdata,)
            _ = _check_signature("precond.setupfn", precond.setupfn, expected)

        expected = (8 + with_userdata,)
        _ = _check_signature("precond.solvefn", precond.solvefn, expected)

    if precond and linsolver in direct:
        raise ValueError("'precond' is not compatitle with direct linear"
                         f" solvers: {direct}.")

    # jactimes
    jactimes = options["jactimes"]
    if jactimes is None:
        pass
    elif not isinstance(jactimes, CVODEJacTimes):
        raise TypeError("'jactimes' must be type CVODEJacTimes.")
    else:

        if jactimes.setupfn:
            expected = (3 + with_userdata,)
            _ = _check_signature("jactimes.setupfn", jactimes.setupfn, expected)

        expected = (5 + with_userdata,)
        _ = _check_signature("jactimes.solvefn", jactimes.solvefn, expected)

    if jactimes and linsolver in direct:
        raise ValueError("'jactimes' is not compatitle with direct linear"
                         f" solvers: {direct}.")
